# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import dataclasses
from .base import NetworkLink, VlanList
'''
    Switch Utilities
    ================

    This Module provides utilities to  ease in converting Carthage :class:`NetworkLink` into switch configuration

'''
__all__ = []


def links_by_port_channel(links: list[NetworkLink]):
    result: dict[str, list[NetworkLink]] = dict()
    for l in links:
        pc_member = getattr(l, 'portchannel_member', None)
        if pc_member:
            result.setdefault(pc_member, [])
            result[pc_member].append(l)
    return result


__all__ += ['links_by_port_channel']


def link_vlan_config(link):
    '''Select a link from which we can get VLAN configuration.

    Try the link itself.  If the link isa member of a bond, try that.
    Also try the same for the other side.  In practice, VLAN
    configuration for bonds probably will not be on the switch side of
    the links because the switch side of a bond link cannot be
    represented when it includes multiple switches.
'''
    def try_link(l):
        if l in already_tried:
            return
        to_try.append(l)
    to_try = [link]
    already_tried = set()
    while to_try:
        link = to_try.pop(0)
        already_tried.add(link)
        if link.allowed_vlans:
            link.allowed_vlans = VlanList.canonicalize(link.allowed_vlans, link)
            if link.untagged_vlan is None and link.net.vlan_id:
                link.untagged_vlan = link.net.vlan_id
            return link
        for member_of_link in link.member_of_links:
            if member_of_link.local_type == 'bond':
                try_link(member_of_link)
        if link.other:
            try_link(link.other)
    return None


__all__ += ['link_vlan_config']


def link_collect_nets(link):
    def try_link(l):
        if l in links:
            return
        to_try.append(l)
    to_try = [link]
    links = set()
    nets = set()
    while to_try:
        link = to_try.pop(0)
        links.add(link)
        if link.net not in nets:
            yield link.net
            nets.add(link.net)
        for l in link.member_of_links:
            try_link(l)
        if link.other:
            try_link(link.other)


def link_collect_vlans(link):
    return {net.vlan_id for net in link_collect_nets(link) if net.vlan_id is not None}


def cisco_vlan_list(vlan_list):
    result = []
    for i in vlan_list:
        if isinstance(i, int):
            result.append(str(i))
        elif isinstance(i, slice):
            result.append(f'{i.start}-{i.end}')
    return ",".join(result)


__all__ += ['cisco_vlan_list']



# Copyright (C) 2018, 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from .schema import ConfigSchema, config_key, ConfigAccessor
from .layout import ConfigLayout
from .types import ConfigPath
from . import base
from ..dependency_injection import *


def inject_config(injector):
    injector.replace_provider(ConfigLayout, allow_multiple=True)


__all__ = ("config_key", "ConfigSchema", "ConfigLayout", "inject_config", 'ConfigAccessor')



# Copyright (C) 2018, 2019, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import collections.abc
import logging

from carthage import *
from carthage.network import this_network, TechnologySpecificNetwork, BridgeNetwork

from .inventory import *
from .utils import wait_for_task
from .connection import VmwareConnection
from .folder import VmwareFolder
from .cluster import VmwareCluster

from pyVmomi import vim

__all__ = (
    'VmwareNetwork',
    'DvSwitch',
    'DistributedPortgroup',
    'our_portgroups_for_switch',
    'NetworkFolder',
    'vmware_trunk_key')


class VmwareNetworkConfig(ConfigSchema, prefix="vmware"):
    distributed_switch: str
    trunk_interface: str


class NetworkFolder(VmwareFolder, kind='network'):
    pass


class DvSwitch(VmwareSpecifiedObject):

    parent_type = NetworkFolder
    stamp_type = "dvswitch"

    def __init__(self, *args, **kwargs):
        if 'name' not in kwargs:
            config = kwargs['config_layout']
            name = config.vmware.distributed_switch
            if name is None:
                raise ValueError(f'unable to create DvSwitch without name')
            if not name.startswith('/'):
                name = f'/{config.vmware.datacenter}/network/{name}'
            kwargs['name'] = name

            kwargs['readonly'] = kwargs.get('readonly', True)
        super().__init__(*args, **kwargs)

    def _config_for(self, host, pnics):

        pspecs = []
        for pnic in pnics:
            pspecs.append(vim.dvs.HostMember.PnicSpec(pnicDevice=pnic.device))

        cs = vim.dvs.HostMember.ConfigSpec()
        cs.operation = vim.ConfigSpecOperation.add
        cs.host = host.mob
        cs.backing = vim.dvs.HostMember.PnicBacking()
        cs.backing.pnicSpec = pspecs
        return cs

    async def update_hosts(self, hostmap):

        hostconfigs = []
        for host, pnics in hostmap.items():
            cs = self._config_for(host, pnics)
            cs.operation = vim.ConfigSpecOperation.add
            for h in self.mob.config.host:
                if host.mob == h.config.host:
                    cs.operation = vim.ConfigSpecOperation.edit
            hostconfigs.append(cs)

        dvspec = vim.DistributedVirtualSwitch.ConfigSpec()
        dvspec.configVersion = self.mob.config.configVersion
        dvspec.host = hostconfigs

        task = self.mob.ReconfigureDvs_Task(dvspec)
        await wait_for_task(task)

    async def do_create(self):

        logging.debug(f'creating distributed portgroup {self.name}')

        cluster = await self.ainjector(VmwareCluster)

        hostconfigs = []

        config = vim.DistributedVirtualSwitch.ConfigSpec()
        config.name = self.name
        config.uplinkPortPolicy = vim.DistributedVirtualSwitch.NameArrayUplinkPortPolicy()
        config.uplinkPortPolicy.uplinkPortName = ['uplink0', 'uplink1']
        config.maxPorts = 8000
        config.host = hostconfigs

        create = vim.DistributedVirtualSwitch.CreateSpec()
        create.configSpec = config
        create.productInfo = vim.dvs.ProductSpec(version='6.6.0')

        task = self.parent.mob.CreateDVS_Task(create)
        await wait_for_task(task)


@inject_autokwargs(network=this_network)
class VmwareNetwork(VmwareNamedObject, TechnologySpecificNetwork):

    '''Abstract Base class representing a VmwareNetwork'''

    def __repr__(self):
        return f"<{self.__class__.__name__} for {self.network.name}: {self.vmware_path}>"


@inject(
    dvswitch=DvSwitch)
class DistributedPortgroup(VmwareNetwork):

    stamp_type = "portgroup"

    parent_type = NetworkFolder

    def __init__(self, *args, config_layout, dvswitch, **kwargs):
        self.dvswitch = dvswitch
        if 'name' not in kwargs:
            kwargs['name'] = kwargs['network'].name
            kwargs['readonly'] = kwargs.get('readonly', False)
            kwargs['parent'] = dvswitch.parent
        super().__init__(*args, **kwargs, config_layout=config_layout)

    @memoproperty
    def name(self):
        return self.network.name

    async def do_create(self):
        logging.debug(f'creating distributed portgroup {self.network.name} on VLAN {self.network.vlan_id}')
        cs = vim.dvs.DistributedVirtualPortgroup.ConfigSpec()
        cs.name = self.full_name
        cs.autoExpand = True
        cs.type = "earlyBinding"
        cs.policy = vim.dvs.VmwareDistributedVirtualSwitch.VMwarePortgroupPolicy()
        cs.policy.trafficFilterOverrideAllowed = True
        cs.policy.macManagementOverrideAllowed = True
        default = vim.dvs.VmwareDistributedVirtualSwitch.VmwarePortConfigPolicy()
        default.macManagementPolicy = vim.dvs.VmwareDistributedVirtualSwitch.MacManagementPolicy()
        default.macManagementPolicy.forgedTransmits = True
        default.macManagementPolicy.allowPromiscuous = True
        default.macManagementPolicy.macChanges = True
        learning = vim.dvs.VmwareDistributedVirtualSwitch.MacLearningPolicy()
        learning.enabled = True
        learning.allowUnicastFlooding = True
        learning.limit = 512
        learning.limitPolicy = "DROP"
        default.macManagementPolicy.macLearningPolicy = learning
        if self.network.vlan_id:
            vlan_id = self.network.vlan_id
            if isinstance(vlan_id, collections.abc.Sequence):
                min, max = vlan_id
                nr = vim.NumericRange()
                nr.start = min
                nr.end = max
                vlan_spec = vim.dvs.VmwareDistributedVirtualSwitch.TrunkVlanSpec()
                vlan_spec.vlanId = [nr]
            else:
                vlan_spec = vim.dvs.VmwareDistributedVirtualSwitch.VlanIdSpec()
                vlan_spec.vlanId = self.network.vlan_id
            default.vlan = vlan_spec
        cs.defaultPortConfig = default
        if not self.mob:
            task = self.dvswitch.mob.AddDVPortgroup_Task([cs])
        else:
            cs.configVersion = self.mob.config.configVersion
            task = self.mob.ReconfigureDVPortgroup_Task(cs)
        await wait_for_task(task)
        try:
            del self.__dict__['mob']
        except KeyError:
            pass

    @memoproperty
    def full_name(self):
        prefix = self.config_layout.container_prefix
        return f"{prefix}{self.network.name}"

    async def delete(self):
        if not self.mob:
            raise RuntimeError(f"{self} does not exist")
        task = self.mob.Destroy()
        await wait_for_task(task)

    async def also_accessed_by(self, others):
        for n in others:
            if isinstance(n, BridgeNetwork):
                trunk = await self._get_trunk()
                ni = trunk.expose_vlan(self.network.vlan_id)
                n.add_member(ni)

    async def _get_trunk(self):
        trunk_base = await self.ainjector.get_instance_async(vmware_trunk_key)
        return await trunk_base.access_by(BridgeNetwork)


@inject(config=ConfigLayout, connection=VmwareConnection)
def our_portgroups_for_switch(switch=None, *, config, connection):
    '''
    :return: Yield  pyVmOmi portgroup objects that match the Carthage prefix for the switch.  The main purpose is for deleting objects.

    :param switch: Name of distributed virtual switch within the Datacenter; if None, uses the switch from the config.

    '''
    if switch is None:
        switch = config.vmware.distributed_switch
    dvs = connection.content.searchIndex.FindByInventoryPath(f"{config.vmware.datacenter}/network/{switch}")
    if dvs is None:
        raise LookupError(f"{switch} DVS not found")
    prefix = config.container_prefix
    for p in dvs.portgroup:
        if p.name.startswith(prefix):
            yield p


@when_needed
@inject(config=ConfigLayout, injector=Injector)
async def _vmware_trunk(config, injector):
    trunk_interface = config.vmware.trunk_interface
    if trunk_interface is None:
        raise ValueError("You must configure config.vmware.trunk_interface")
    # This is a bit hackish.  Our goal is to construct a BridgeNetwork
    # with the right interface name to keep track of vlan interfaces
    # so they can be deleted.
    net = injector(Network, "Vmware Trunk")
    bridge = net.ainjector.injector(BridgeNetwork, bridge_name=trunk_interface, delete_bridge=False)
    net.ainjector.add_provider(bridge)
    return net

vmware_trunk_key = InjectionKey(Network, vmware_role='trunk')



# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import dataclasses
import typing
from ..dependency_injection import *
from .base import NetworkInterface, logger, NetworkLink, BridgeNetwork, VethInterface
from .. import sh


@dataclasses.dataclass
class NetworkNamespace:

    name: str
    network_links: typing.Dict[str, NetworkLink]

    def __post_init__(self):
        logger.debug("Bringing up network namespace for %s", self.name)
        try:
            sh.ip(
                "netns", "add",
                self.name)
        except sh.ErrorReturnCode_1:  # link exists
            sh.ip("netns", "delete", self.name)
            sh.ip("netns", "add", self.name)
        self.closed = False

    async def start_networking(self):
        for interface, link in self.network_links.items():
            net = await link.instantiate(BridgeNetwork)
            veth = net.add_veth(link, self)

    def close(self):
        if self.closed:
            return
        self.closed = True
        logger.info("Deleting network namespace %s", self.name)
        try:
            sh.ip("netns", "delete", self.name)
        except sh.ErrorReturnCode:
            logger.exception("Error deleting network namespace")



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *

import asyncio
import logging
import time
from pathlib import Path
import carthage.network

from .image import VmwareDataStore, VmdkTemplate, vm_datastore_key
from .inventory import VmwareNamedObject, VmwareSpecifiedObject
from .connection import VmwareConnection
from .folder import VmwareFolder
from .cluster import VmwareCluster
from .datacenter import VmwareDatacenter
from .utils import wait_for_task
from . import network
from .config_spec import *

from pyVmomi import vim, vmodl

logger = logging.getLogger('carthage.vmware.vm')


class VmwareVmConfig(ConfigSchema, prefix="vmware"):
    datacenter: str
    folder: str = "carthage"
    cluster: str


class HardwareConfig(ConfigSchema, prefix="vmware.hardware"):
    boot_firmware: str = "efi"
    version: int = 14
    memory_mb: int = 4096
    disk: int = 25000000000
    cpus: int = 1
    paravirt: bool = True


vmware_config = config_key('vmware')


class VmFolder(VmwareFolder, kind='vm'):

    def __init__(self, name=None, **kwargs):
        if name is None:
            name = kwargs['config_layout'].vmware.folder
        kwargs['name'] = name
        kwargs.setdefault('readonly', False)
        super().__init__(**kwargs)

    @memoproperty
    def inventory_view(self):
        return self.injector(VmInventory, folder=self)

    async def delete(self, cluster=None):
        v = None
        if self.mob is None:
            return
        try:
            v = self.children([vim.VirtualMachine])
            cluster_name = self.config_layout.vmware.cluster
            if not cluster:
                datacenter = await self.ainjector(VmwareDatacenter)
                cluster = await self.ainjector(
                    VmwareCluster, name=cluster_name,
                    parent=datacenter.host_folder if not cluster_name.startswith('/') else None)

                rp = cluster.mob.resourcePool
            for vm in v:
                try:
                    vm.MarkAsVirtualMachine(rp)
                except BaseException:
                    pass
                try:
                    vm.PowerOff()
                except BaseException:
                    pass
        finally:
            if v:
                v.close()
        await asyncio.sleep(2)
        return await wait_for_task(self.mob.Destroy())


@inject(
    storage=vm_datastore_key,
    cluster=VmwareCluster,
    parent=InjectionKey(VmFolder, optional=True),
    network_config=InjectionKey(carthage.network.NetworkConfig, optional=True),
    # We add a template VM to the dependencies later avoiding a forward reference
)
class Vm(VmwareSpecifiedObject, Machine):

    stamp_type = "vm"
    parent_type = VmFolder
    config_spec_class = vim.vm.ConfigSpec

    nested_virt = False  # : Enable nested virtualization

    def __init__(self, name, template,
                 guest_id="ubuntu64Guest",
                 *, storage, cluster, dspath=None, network_config=None,
                 console_needed=True, **kwargs):
        kwargs.setdefault('readonly', False)
        super().__init__(name=name, **kwargs)

        self.cluster = cluster
        self.storage = storage
        if dspath is None:
            self.dspath = self.config_layout.vmware.datastore.path
        else:
            self.dspath = dspath
        if self.dspath and not self.dspath.endswith('/'):
            self.dspath += '/'
        self.running = False
        self.folder = self.parent
        vm_config = self.config_layout.vmware
        self.cpus = vm_config.hardware.cpus
        self.memory = vm_config.hardware.memory_mb
        self.paravirt = vm_config.hardware.paravirt
        self.disk_size = vm_config.hardware.disk
        if self.config_layout.vm_image_size > self.disk_size:
            self.disk_size = int(self.config_layout.vm_image_size)
        self.guest_id = guest_id
        self.network_config = network_config
        self.template = None
        self.template_snapshot = None
        self.inventory_object = None
        if template:
            self.template = template
            self.template_snapshot = getattr(template, 'clone_from_snapshot', None)
        self._operation_lock = asyncio.Lock()

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"

    @classmethod
    def satisfies_injection_key(cls, k):
        if issubclass(k.target, Vm):
            if set(k.constraints) - {'role'} == set():
                return True
        return super().satisfies_injection_key(k)

    async def delete(self):
        try:
            task = self.mob.PowerOffVM_Task()
            await carthage.vmware.utils.wait_for_task(task)
        except (vim.fault.InvalidPowerState, vmodl.fault.NotSupported):
            pass
        task = self.mob.Destroy_Task()
        await carthage.vmware.utils.wait_for_task(task)
        self.mob = None
        self.vmware_uuid = None

    async def build_config(self, mode, oconfig=None, stagefilter=None):
        config = await super().build_config(mode, oconfig=oconfig, stagefilter=stagefilter)
        cur = -1
        for c in config.deviceChange:
            if c.device.key == 0:
                assert cur > -100
                c.device.key = cur
                cur = cur - 1
        return config

    async def do_create(self):
        try:
            if self.template:
                await self.template.async_become_ready()
                spec = vim.vm.CloneSpec()
                locspec = vim.vm.RelocateSpec()
                spec.location = locspec
                spec.config = await self.build_config('clone', oconfig=self.template.mob.config)
                locspec.datastore = self.storage.mob
                locspec.pool = self.cluster.mob.resourcePool
                if self.template_snapshot:
                    for s in self.template.all_snapshots():
                        if s.name == self.template_snapshot:
                            spec.snapshot = s.snapshot
                            locspec.diskMoveType = "createNewChildDiskBacking"
                            break

                res = self.template.mob.Clone(
                    self.parent.mob,
                    self.full_name,
                    spec,
                )
            else:
                config = await self.build_config('create')
                res = self.parent.mob.CreateVm(config, self.cluster.mob.resourcePool)
            await carthage.vmware.utils.wait_for_task(res)

        except carthage.ansible.AnsibleFailure as e:
            if "is not supported" in e.ansible_result.tasks['vmware_guest'].msg:
                # Everything but customization worked
                res = e.ansible_result
            else:
                raise

            self.mob = self._find_from_path()
        return res

    async def start_machine(self):
        loop = self.injector.get_instance(asyncio.AbstractEventLoop)
        async with self._operation_lock:
            if not self.running:
                logger.debug(f'Starting dependencies for {self.name}')
                await self.start_dependencies()
                logger.info(f'Starting {self.name} VM')
                power = self.mob.summary.runtime.powerState
                if power != "poweredOn":
                    task = self.mob.PowerOn()
                    await wait_for_task(task)
                    if task.info.state != 'success':
                        raise RuntimeError(task.info.error)
                self.running = True
            if self.__class__.ip_address is Machine.ip_address:
                try:
                    self.ip_address
                except NotImplementedError:
                    await loop.run_in_executor(None, self._get_ip_address)
            return True

    async def stop_machine(self):
        async with self._operation_lock:
            if not self.running:
                return
            logger.info(f'Stopping {self.name} VM')
            power = self.mob.summary.runtime.powerState
            if power == "poweredOn":
                task = self.mob.PowerOff()
                await wait_for_task(task)
            self.running = False
            return True

    def _get_ip_address(self):
        for i in range(20):
            try:
                nets = self.mob.guest.net
                nets.sort(key=lambda x: x.deviceConfigId)
                ip = nets[0].ipAddress[0]
                self.ip_address = ip
                return
            except IndexError:
                pass
            time.sleep(5)
        raise TimeoutError(f'Unable to get IP address for {self.name}')

    def all_snapshots(self):
        sn = list()

        def recurse(s):
            sn.append(s)
            for c in s.childSnapshotList:
                recurse(c)
        for s in self.mob.snapshot.rootSnapshotList:
            recurse(s)
        return sn


@inject(template=None)
class VmTemplate(Vm):

    clone_from_snapshot = "template_snapshot"

    def __init__(self, name, disk="unspecified", **kwargs):
        if disk == "unspecified" and not kwargs.get('template', None):
            raise TypeError(
                "If you want a VM Template with no disk and no parent explicitly request that by setting disk to None")
        elif disk == "unspecified":
            disk = None
        self.disk = disk
        if disk:
            kwargs['template'] = None
            self.clone_from_snapshot = None  # Doesn't tend to work with explicit disks
        super().__init__(name=name, **kwargs)
        self.network_config = None

    @setup_task("create_clone_snapshot")
    @inject(loop=asyncio.AbstractEventLoop)
    async def create_clone_snapshot(self, loop):
        if self.clone_from_snapshot is None:
            raise SkipSetupTask
        t = self.mob.CreateSnapshot("template_snapshot", "Snapshot for template clones", False, True)
        await wait_for_task(t)

    @setup_task("Mark as template")
    async def mark_as_template(self):
        # ESXI 6.7 seems to have a bug  producing linked clones from templates rather than VMs
        if self.clone_from_snapshot:
            raise SkipSetupTask
        try:
            self.mob.MarkAsTemplate()
        except Exception:
            pass


class BasicConfig(ConfigSpecStage, stage_for=Vm, order=20,
                  mode=('create', 'reconfig', 'clone')):

    def apply_config(self, config):
        obj = self.obj
        vmc = self.obj.config_layout.vmware
        if not obj.mob:
            dsname = obj.storage.name
            config.files = vim.vm.FileInfo(vmPathName=f'[{dsname}]{obj.dspath}')
        if self.bag.mode == 'create':
            config.name = obj.full_name
        config.numCPUs = obj.cpus
        config.memoryMB = obj.memory
        config.nestedHVEnabled = obj.nested_virt
        config.firmware = vmc.hardware.boot_firmware
        config.version = "vmx-" + str(vmc.hardware.version)
        config.guestId = obj.guest_id


class ScsiSpec(DeviceSpecStage, stage_for=Vm,
               order=30,
               dev_classes=(vim.vm.device.ParaVirtualSCSIController,
                            vim.vm.device.VirtualLsiLogicSASController,
                            vim.vm.device.VirtualLsiLogicController,
                            vim.vm.device.VirtualBusLogicController)):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.obj.paravirt:
            self.expected_dev = self.dev_classes[0]
        else:
            self.expected_dev = self.dev_classes[1]

    def filter_device(self, d):
        if isinstance(d, self.expected_dev):
            self.bag.scsi_key = d.key
            return True
        return False

    def new_devices(self, config):
        if getattr(self.bag, 'scsi_key', None) is None:
            d = self.expected_dev(busNumber=0,
                                  sharedBus="noSharing",
                                  key=-100)
            self.bag.scsi_key = -100
            return [d]
        return []


class DiskSpec(DeviceSpecStage,
               stage_for=Vm,
               order=120,
               dev_classes=vim.vm.device.VirtualDisk):

    disk_found = False  # Set in instances once a disk is found

    def filter_device(self, d):
        changed = False
        self.disk_found = True

        if self.bag.mode == 'clone' and d.backing.thinProvisioned is False:
            d.backing.thinProvisioned = True
            changed = True

        if False:
            if self.bag.mode == "clone":
                return d if changed else True

        if d.capacityInBytes < self.obj.disk_size:
            if (self.bag.mode == "reconfig" and d.backing.parent) or (
                    self.bag.mode == "clone" and self.obj.template_snapshot):
                raise ValueError("You cannot increase the capacity of a disk with a parent backing")
            d.capacityInBytes = self.obj.disk_size
            d.capacityInKB = int(d.capacityInBytes / 1024)
            changed = True

        if d.controllerKey != self.bag.scsi_key:
            d.controllerKey = self.bag.scsi_key
            changed = True
        return d if changed else True

    async def new_devices(self, config):
        if self.disk_found:
            return []
        d = self.dev_classes[0]()
        d.backing = vim.vm.device.VirtualDisk.FlatVer2BackingInfo()
        d.controllerKey = self.bag.scsi_key
        d.backing.thinProvisioned = True
        d.backing.diskMode = 'persistent'
        orig_disk = getattr(self.obj, 'disk', None)
        if orig_disk:
            if not isinstance(orig_disk, VmdkTemplate):
                orig_disk = await self.obj.ainjector(orig_disk)
            await orig_disk.async_become_ready()
            d.backing.fileName = orig_disk.disk_path
        d.capacityInBytes = self.obj.disk_size
        d.capacityInKB = int(d.capacityInBytes / 1024)
        d.unitNumber = 0
        if not orig_disk:
            self.file_operation = 'create'
        return [d]


class NetSpecStage(DeviceSpecStage,
                   stage_for=Vm,
                   order=110,
                   dev_classes=(vim.vm.device.VirtualE1000, vim.vm.device.VirtualVmxnet3),
                   ):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.obj.paravirt:
            self.expected_dev = vim.vm.device.VirtualVmxnet3
        else:
            self.expected_dev = vim.vm.device.VirtualE1000

        self.net_index = 0

    async def apply_config(self, config):
        await self.obj.resolve_networking()
        if not self.obj.network_links:
            return []
        self.network_links = list(self.obj.network_links.values())
        for n in self.network_links:
            await n.instantiate(network.DistributedPortgroup)
        await self.obj.ainjector(super().apply_config, config)

    async def new_devices(self, config):
        devs = []
        links = self.network_links[self.net_index:]
        for link in links:
            net = link.net_instance
            mac = link.mac
            if not net:
                continue
            pc = vim.dvs.PortConnection()
            pc.portgroupKey = net.mob.key
            pc.switchUuid = net.mob.config.distributedVirtualSwitch.uuid

            ds = self.expected_dev()
            ds.backing = vim.vm.device.VirtualEthernetCard.DistributedVirtualPortBackingInfo()
            ds.backing.port = pc
            if mac:
                ds.addressType = 'manual'
                ds.macAddress = mac
            devs.append(ds)
        return devs

    def filter_device(self, d):
        if self.bag.mode == 'clone':
            return False
        link = self.network_links[self.net_index]
        net = link.net_instance
        mac = link.mac
        if d.backing.port.portgroupKey != net.mob.key:
            return False
        self.net_index += 1
        return True


@inject(folder=VmwareFolder, connection=VmwareConnection)
class VmInventory(Injectable):

    def __init__(self, *, folder, connection):
        self.view = connection.content.viewManager.CreateContainerView(
            folder.mob,
            [vim.VirtualMachine], True)

    def find_by_name(self, name):
        for v in self.view.view:
            if v.name == name:
                return v
        return None

    def close(self):
        try:
            self.view.Destroy()
            del self.view
        except BaseException:
            pass


# Now mark VM as taking a template A vm can be a template for a vm, but
# we want to look up with a role for finding in the injector
inject(template=InjectionKey(Vm, role='template', _optional=True, _ready=False))(Vm)

from carthage.vm import vm_image


@inject(image=vm_image)
class vm_template_from_image(AsyncInjectable):

    @classmethod
    def default_class_injection_key(cls):
        return InjectionKey(Vm, role='template')

    async def async_resolve(self):
        vmdk = await self.ainjector(VmdkTemplate, image=self.image)
        return await self.ainjector(VmTemplate, name=(Path(self.image.path).stem) + ".template", disk=vmdk)


from . import inventory



# Copyright (C) 2018, 2019, 2020, 2021, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import typing


def combine_mro(
        base: typing.Union[type, typing.Sequence[type]],
        subclass: type, attribute: str,
        add: typing.Callable,
        state):
    # for all members of the mro of base that are subclasses of
    # *subclass*, run ``add(mro_member, getattr(mro_member,
    # attribute), state)`` *base* may be a sequence This is intended
    # for collecting together attributes like classes_to_inject or
    # namespace_filters that are manually set on a class and need to
    # be combined together.  Currently, if there are conflicts (as in
    # multiple classes setting the same key in a mapping), the
    # visitation order is wrong and the wrong value will survive.
    #
    # This function is not appropriate for values like
    # __initial_injections__, __container_propagations__, or
    # __transclusions__ that are automatically maintained and built up
    # as the inheritance tree is constructed.  For one thing, it is
    # only necessary to descend one level since the earlier levels
    # should have already accumulated together.  Also, typically,
    # those attributes are attributes of say ModelingContainer or
    # InjectableModelType instances rather than of say
    # ModelingContainer subclasses, and so the subclass check is
    # guaranteed to fail.
    if isinstance(base, type):
        base = [base]
        mro_set = set()
        mro = []
        for b in base:
            if b not in mro_set and issubclass(b, subclass):
                mro_set.add(b)
                mro.append(b)
                try:
                    res = getattr(b, attribute)
                    add(b, res, state)
                except AttributeError:
                    pass
            for m in b.__mro__:
                if m in mro_set:
                    continue
                if not issubclass(m, subclass):
                    continue
                mro_set.add(m)
                mro.append(m)
                try:
                    res = getattr(m, attribute)
                except AttributeError:
                    continue
                add(m, res, state)


def combine_mro_list(base, subclass, attribute):
    def add(m: type, res: list, state):
        for l in res:
            if l not in state:
                state.append(l)
    state = []
    combine_mro(base, subclass, attribute, add, state)
    return state


def combine_mro_mapping(base, subclass, attribute) -> typing.Dict[str, typing.Any]:
    def add(m, res, state):
        for k, v in res.items():
            if k not in state:
                state[k] = v
    state: typing.Dict[str, typing.Any] = {}
    combine_mro(base, subclass, attribute, add, state)
    return state


__all__ = [
    'combine_mro_list',
    'combine_mro_mapping'
]


def setattr_default(obj, a: str, default, inherited_ok=False):
    if inherited_ok:
        has_attr = hasattr(obj, a)
    else:
        has_attr = a in obj.__dict__
    if not has_attr:
        setattr(obj, a, default)


__all__ += ['setattr_default']


def gather_from_class(self, *keys, mangle_name=True):
    '''
    :param mangle_name: If true, and name is not in class, set name from __name__
    '''

    d: dict = {}
    if isinstance(self, type):
        cls = self
    else:
        cls = self.__class__
    for k in keys:
        try:
            d[k] = getattr(cls, k)
        except AttributeError:
            if k == 'name' and mangle_name:
                d['name'] = cls.__name__.lower()
    return d


__all__ += ['gather_from_class']


def key_from_injector_access(*accesses):
    from .decorators import injector_access
    result = []
    for k in accesses:
        if isinstance(k, injector_access):
            k = k.key
        elif  hasattr(k, '__provides_dependencies_for__'):
            k = k.__provides_dependencies_for__[0]
        result.append(k)
    return result


__all__ += ['key_from_injector_access']


def fixup_dynamic_name(n):
    return n.replace('-', '_')


__all__ += ['fixup_dynamic_name']



# Copyright (C) 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.


from pathlib import Path
import pkg_resources
import typing
from .dependency_injection import inject, Injector, AsyncInjector
from .plugins import CarthagePlugin
from . import sh
__all__ = []

dependency_preference_order = ['deb', 'pypi']

@inject(injector=Injector)
def plugin_iterator(injector):
    for key, plugin in injector.filter_instantiate(
            CarthagePlugin, ['name']):
        yield plugin

__all__ += ['plugin_iterator']

def collect_dependencies(
        plugins: list[CarthagePlugin],
        descend_plugins: typing.Union[list,bool]=[],
        minimize_os_packages:bool = True,
        valid_dependency_types = dependency_preference_order,
) -> list[dict[str, str]]:
    '''
    Iterate over the plugins.  Return dependencies of the form ``{type:dependency}`` for each dependency in a plugin.  Dependencies may be expressed multiple times.

    :param descend_plugins: A list of  plugins where even if the plugin is installed as a python distribution, we should process pypi dependencies.  By default, if a plugin is installed as a distribution, we assume that some OS or PyPi installation has taken care of installing its Python requirements.  (We may need to do something similar for OS level requirements if plugins start being installed as OS packages).  This is a list of plugins where we return all their dependencies even if they are installed as python distributions.  The primary case where this is used is so that the list of dependencies for plugins bundled with Carthage core can be generated or so a plugin's pyproject.toml can be updated from its Carthage metadata.

    :param minimize_os_packages: If a dependency is available both as an OS package and a PyPI distribution, we typically prefer the OS package.  However, at least when installing packages, there is no need to install an OS package if the distribution is already available either via OS installation or PyPi.  If this parameter is true, skip dependencies for any dependency having a PyPi satisfaction that is already installed.  This only makes sense for installs, and not for generating dependency lists.

    '''
    results: list[dict] = []
    for plugin in plugins:
        if not 'dependencies' in plugin.metadata: continue
        skip_pypi_package = False # Set per plugin based on descend_plugins
        if 'package' in plugin.metadata:
            if not ( descend_plugins is True or plugin.name in descend_plugins):
                try:
                    pkg_resources.get_distribution(plugin.metadata['package'])
                    skip_pypi_package = True
                except pkg_resources.ResolutionError: pass

        for dependency in plugin.metadata['dependencies']:
            assert isinstance(dependency, dict)
            skip_os_package = minimize_os_packages and 'pypi' in dependency # set to False below; decided per dependency
            if minimize_os_packages and 'pypi' in dependency:
                for requirement in pkg_resources.parse_requirements(dependency['pypi']):
                    # We only expect one requirement but the interface returns an iterator
                    # This code would be better if it checked the installed version satisfied the requirement
                    try: pkg_resources.get_distribution(requirement.name)
                    except pkg_resources.ResolutionError:
                        skip_os_package = False
                        break
            for dependency_type in valid_dependency_types:
                # we return the first dependency that is valid,
                # unless we choose to skip that dependency.  If we
                # skip a dependency, no dependency is returned; we
                # skip when none is needed.
                if dependency_type in dependency: break
            else: # This else needs to belong to the for loop not the if
                dependency_type = None
            if dependency_type == 'deb':
                if skip_os_package: continue
                results.append(dict(deb=dependency['deb']))
            elif dependency_type == 'pypi':
                if skip_pypi_package: continue
                results.append(dict(pypi=dependency['pypi']))
    return results

__all__ += ['collect_dependencies']

@inject(injector=Injector)
async def install_deb(*, injector, **kwargs):
    plugins = list(injector(plugin_iterator))
    # Foreground and stem probably become configurable if we want to
    # be able to turn off sudo or install remotely
    foreground = True
    stem = sh
    try: stem = stem.sudo.bake()
    except sh.CommandNotFound: pass
    packages:list[str] = []
    for dependency in collect_dependencies(plugins, **kwargs):
        if 'deb' in dependency:
            packages.append(dependency['deb'])
            # if not a deb dependency then deb is not the most
            # preferred way to install; our caller could have adjusted
            # valid_dependency_types if they wanted deb more
            # preferred.
    if packages:
        stem.apt(
            '-y', '--no-install-recommends',
            'install', *packages, _fg=foreground)

@inject(injector=Injector)
async def install_pypi(*, sudo=True,
                       allow_system=True,
                       injector, **kwargs):
    stem = sh
    foreground = True
    if sudo:
        try: stem = getattr(stem, 'sudo').bake()
        except sh.CommandNotFound: pass
    requirements = []
    for dependency in collect_dependencies(plugins, **kwargs):
        if 'pypi' in dependency: requirements.append(dependency['pypi'])
    options = []
    if allow_system: options.append('--break-system-packages')
    if requirements:
        await stem.pip3(
            'install',
            *options,
            *requirements,
            _fg=foreground)

@inject(ainjector=AsyncInjector)
async def install_carthage_dependencies(
        minimize_os_packages:bool = False,
        valid_dependency_types:list = dependency_preference_order,
        *, ainjector):
    await ainjector(
        install_deb,
        minimize_os_packages=minimize_os_packages,
        valid_dependency_types=valid_dependency_types)
    
__all__ += ['install_carthage_dependencies']


@inject(injector=Injector)
def gen_requirements(descend_plugins=[],
                     *,
                     injector):
    plugins = list(injector(plugin_iterator))
    requirements = []
    for dependency in collect_dependencies(plugins,
                                           valid_dependency_types=['pypi'],
                                           minimize_os_packages=False,
                                           descend_plugins=descend_plugins):
        requirements.append(dependency['pypi'])
    return requirements

__all__ += ['gen_requirements']

@inject(injector=Injector)
def gen_os_dependencies(package_type: str = 'deb',
                        minimize_os_packages: bool=False,
                     *,
                     injector):
    plugins = list(injector(plugin_iterator))
    packages = []
    for dependency in collect_dependencies(
            plugins,
            valid_dependency_types=[package_type],
            minimize_os_packages=minimize_os_packages,
                                           ):
        packages.append(dependency[package_type].strip())
    return packages

__all__ += ['gen_os_dependencies']

@inject(injector=Injector)
def gen_requirements_command(args, injector):
    print("\n".join(injector(gen_requirements, args.descend_plugins)))

__all__ += ['gen_requirements_command']

@inject(ainjector=AsyncInjector)
async def install_carthage_dependencies_command(args, ainjector):
    valid_dependency_types = dependency_preference_order
    if args.prefer_python:
        valid_dependency_types = ['pypi', 'deb']
    await ainjector(
        install_carthage_dependencies,
        minimize_os_packages=args.minimize_os,
        valid_dependency_types=valid_dependency_types)

__all__ += ['install_carthage_dependencies_command']

def setup_deployment_commands(command_action):
    gen_requirements = command_action.add_parser(
        'generate_requirements',
        help='Generate a set of Python requirements either for a requirements.txt or for use in pyproject.toml')
    gen_requirements.add_argument(
        '--descend-plugins',
        nargs='+',
        default=[],
        help='Plugins whose internal dependencies should be included'
        )
    install_dependencies = command_action.add_parser(
        'install_dependencies', 
        help='Install Carthage dependencies')
    install_dependencies.add_argument(
        '--no-minimize-os',
        dest='minimize_os',
        action='store_false',
        default=True,
        help='Do not minimize OS package installs when a Python distribution is already installed')
    install_dependencies.add_argument(
        '--prefer-python',
        action='store_true',
        help='Prefer pypi to OS packages')
    
    
__all__ += ['setup_deployment_commands']



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *

from .inventory import VmwareNamedObject
from .folder import HostFolder


class VmwareCluster(VmwareNamedObject, kind='cluster'):

    parent_type = HostFolder

    def __init__(self, name=None, parent=None, mob=None, **kwargs):
        if not mob and not parent and not name:
            vmc = kwargs['config_layout'].vmware
            parent = f'/{vmc.datacenter}/host'
            name = vmc.cluster
        self.spec = kwargs.pop('spec', None)
        super().__init__(name=name, parent=parent, mob=mob, **kwargs)

    async def do_create(self):
        self.mob = self.parent.mob.CreateClusterEx(self.name, self.spec)
        return self.mob



# Copyright (C) 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage.config import ConfigSchema
from carthage.config.types import ConfigPath
from carthage.dependency_injection import *
from carthage.machine import customization_task


class ExtraPackagesConfig(ConfigSchema, prefix="extra_packages"):

    architectures: str = "amd64 source"
    components: str = "main contrib non-free proprietary"

    #: Accept any .changes file from this directory
    packages_dir: ConfigPath

    repository_dir: ConfigPath = "{state_dir}/extra_packages_repo"


@inject(injector=Injector)
def carthage_plugin(injector):
    from .repo import ExtraRepository, AddExtraRepo, UpgradeFromRepo
    from carthage.hadron.images import HadronImageMixin
    from carthage.hadron.build_database import RouterMixin, NonRouterMixin
    HadronImageMixin.add_extra_repo = customization_task(
        AddExtraRepo,
        before=HadronImageMixin.setup_hadron_packages)
    for c in (NonRouterMixin, RouterMixin):
        c.upgrade_from_extra_repo = customization_task(
            UpgradeFromRepo, before=RouterMixin.ansible_initial_router)
        c.add_extra_repo = customization_task(
            AddExtraRepo,
            before=c.upgrade_from_extra_repo)

    injector.add_provider(ExtraRepository)



# Copyright (C) 2019, 2020, 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import asyncio
import contextlib
import dataclasses
import datetime
import logging
import os
import os.path
import time
import typing
import sys
import shutil
import weakref
import importlib.resources
from pathlib import Path
import carthage
from carthage.dependency_injection import AsyncInjector, inject, BaseInstantiationContext, InjectionKey
from carthage.dependency_injection.introspection import current_instantiation
from carthage.config import ConfigLayout
from carthage.utils import memoproperty, import_resources_files
import collections.abc

__all__ = ['logger', 'TaskWrapper', 'TaskMethod', 'setup_task', 'SkipSetupTask', 'SetupTaskMixin',
           'cross_object_dependency',
           'mako_task',
           'install_mako_task']

logger = logging.getLogger('carthage.setup_tasks')

_task_order = 0


def _inc_task_order():
    global _task_order
    t = _task_order
    _task_order += 100
    return t


class SetupTaskContext(BaseInstantiationContext):

    def __init__(self, instance, task):
        super().__init__(instance.injector)
        self.instance = instance
        self.task = task

    def __enter__(self):
        res = super().__enter__()
        if self.parent:
            self.parent.dependency_progress(self.task.stamp, self)
        return res

    def done(self):
        if self.parent:
            self.parent.dependency_final(self.task.stamp, self)
        super().done()

    @property
    def description(self):
        return f'setup_task: {self.instance}.{self.task.stamp}'

    def get_dependencies(self):
        from .dependency_injection.introspection import get_dependencies_for
        return get_dependencies_for(self.task, self.instance.injector)

@dataclasses.dataclass
class TaskInspector:

    task: TaskWrapperBase
    #: Either the class or instance with the task
    from_obj: typing.Union[SetupTaskMixin, type[SetupTaskMixin]]
    #: The task previous in the dependency chain.
    previous: typing.Optional[TaskInspector] = dataclasses.field(default=None, repr=False)

    @inject(ainjector=AsyncInjector)
    async def should_run(self, ainjector):
        if not self.is_instance:
            raise NotImplementedError('You must inspect an instantiated class to determine if a task will run.')
        dependency_last_run = await self.dependency_last_run(ainjector)
        should_run, last_run = await  self.task.should_run_task(self.from_obj, ainjector=ainjector, run_methods=False)
        self.last_run = last_run
        return should_run

    @property
    def stamp(self):
        "The complete stamp that will be checked; for customizations includes the stamp_stem"
        try: return self._stamp
        except AttributeError:
            return self.task.stamp

    @stamp.setter
    def stamp(self, val):
        self._stamp = val
        

    @property
    def instance_id(self):
        "The ID of the instance on which this task is run; for Customizations, the ultimate host's ID, **not** the id of the instantiated :class:`carthage.machine.BaseCustomization`"
        try: return self._instance_id
        except AttributeError:
            return id(self.from_obj)

    @instance_id.setter
    def instance_id(self, val):
        self._instance_id = val
        
    def subtasks(self):
        from .machine import CustomizationWrapper
        if isinstance(self.task, CustomizationWrapper):
            return self.task.inspect(self.from_obj, instance_id=self.instance_id)
        return []
    
    async def dependency_last_run(self, ainjector):
        try: return self._dependency_last_run
        except AttributeError: pass
        inspector_list = []
        current = self.previous
        dependency_last_run = None
        while current:
            try:
                dependency_last_run = inspector.last_run
                break
            except AttributeError:pass
            inspector_list.insert(0,current)
            current = current.previous
        for inspector in inspector_list:
            if dependency_last_run and not hasattr(inspector, '_dependency_last_run'):
                inspector._dependency_last_run = dependency_last_run
            await inspector.should_run(ainjector)
            dependency_last_run = inspector.last_run
        if dependency_last_run is None:
            # Happens when no previous dependencies
            dependency_last_run = 0.0
        self._dependency_last_run = dependency_last_run
        return dependency_last_run

    @property
    def description(self):
        return f'{self.task.description} for {self.from_obj}'

    @property
    def is_instance(self):
        '''Return true  if this is inspecting an instance  that has been instantiated rather than inspecting tasks on a class.
        '''
        from .machine import CustomizationInspectorProxy
        return isinstance(self.from_obj, (SetupTaskMixin, CustomizationInspectorProxy))

                
    

@dataclasses.dataclass
class TaskWrapperBase:

    description: str
    order: int = dataclasses.field(default_factory=_inc_task_order)
    invalidator_func = None
    check_completed_func = None
    hash_func: typing.Callable = staticmethod(lambda self: "")

    @memoproperty
    def stamp(self):
        raise NotImplementedError

    def __set_name__(self, owner, name):
        self.stamp = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return TaskMethod(self, instance)

    def __call__(self, instance, *args, **kwargs):
        def success():
            if not self.check_completed_func:
                hash_contents = instance.injector(self.hash_func, instance)
                instance.create_stamp(self.stamp, hash_contents)
            instance.injector.emit_event(
                InjectionKey(SetupTaskMixin), "task_run",
                instance, task=self,
                adl_keys=instance.setup_task_event_keys())
            
        def fail(e):
            if not self.check_completed_func:
                instance.logger_for().warning(
                    f'Deleting {self.description} task stamp for {instance} because task failed')
                instance.delete_stamp(self.stamp)
            instance.injector.emit_event(
                InjectionKey(SetupTaskMixin), "task_fail",
                instance, task=self,
                exception=e,
                adl_keys=instance.setup_task_event_keys())
    
        def final():
            context.done()

        def callback(fut):
            exc = fut.exception()
            if exc is None:
                success()
            elif isinstance(exc,SkipSetupTask):
                pass
            else:
                fail(exc)
            final()
        mark_context_done = True
        with contextlib.ExitStack() as stack:
            context = current_instantiation()
            if isinstance(context, SetupTaskContext) \
               and context.instance is instance and context.task is self:
                pass  # This is the right context to use; set up by run_setup_tasks
            else:
                context = SetupTaskContext(instance, self)
                stack.enter_context(context)
            try:
                res = self.func(instance, *args, **kwargs)
                if isinstance(res, collections.abc.Coroutine):
                    res = instance.ainjector.loop.create_task(res)
                    res.add_done_callback(callback)
                    mark_context_done = False
                    instance.injector.emit_event(
                        InjectionKey(SetupTaskMixin), "task_start",
                        instance, task=self,
                        adl_keys=instance.setup_task_event_keys())
                    if hasattr(instance, 'name'):
                        res.purpose = f'setup task: {self.stamp} for {instance.name}'
                    return res
                else:
                    success()
                    return res
            except SkipSetupTask:
                raise
            except Exception as e:
                fail(e)
                raise
            finally:
                if mark_context_done:
                    final()

    async def should_run_task(self, obj: SetupTaskMixin,
                              dependency_last_run: float = None,
                              *, ainjector: AsyncInjector, run_methods=True):

        '''Indicate whether this task should be run for *obj*.

        :returns: Tuple of whether the task should be run and when the task was last run if ever.

        * If :meth:`check_completed` has been called, then the task should be run when either the check_completed function returns falsy or our dependencies have been run more recently.

        * Otherwise, if there is no stamp then this task should run

        * If there is a :meth:`invalidator`, then this task should run if the invalidator returns falsy.

        * This task should run if any dependencies have run more recently than the stamp

        * If there is a nontrivial hash_func, then the hash contents
          stored in the stamp are compared to the results of the
          hash_func.  If the Contents differ, this task should run.

        * Otherwise this task should not run.

        :param: obj
            The instance on which setup_tasks are being run.
        :param run_methods: If False , no check_completed or invalidators are run.  If the task has a check_completed function, return None rather than True or False for wether the task should run.  Invalidators and hash functions are ignored, and the return is dependent only on stamps.

        '''
        if dependency_last_run is None:
            dependency_last_run = 0.0
        if self.check_completed_func:
            if not run_methods: return (None, dependency_last_run)
            last_run = await ainjector(self.check_completed_func, obj)
            hash_contents = ""
            if last_run is True:
                obj.logger_for().debug(f"Task {self.description} for {obj} determined complete by check_completed_func(); no timestamp provided")
                return (False, dependency_last_run)
        else:
            last_run, hash_contents = obj.check_stamp(self.stamp)
        if last_run is False:
            obj.logger_for().debug(f"Task {self.description} never run for {obj}")
            return (True, dependency_last_run)
        if last_run < dependency_last_run:
            obj.logger_for().debug(
                f"Task {self.description} last run {_iso_time(last_run)}, but dependency run more recently at {_iso_time(dependency_last_run)}")
            return (True, dependency_last_run)
        if (not self.check_completed_func) and run_methods:
            actual_hash_contents = await ainjector(self.hash_func, obj)
            if actual_hash_contents != hash_contents:
                obj.logger_for().info(f'Task {self.description} invalidated by hash_func() change from `{hash_contents}` to `{actual_hash_contents}`; last run {_iso_time(last_run)}')
                return (True, dependency_last_run)
        if self.invalidator_func and run_methods:
            if not await ainjector(self.invalidator_func, obj, last_run=last_run):
                obj.logger_for().info(f"Task {self.description} invalidated for {obj} by invalidator_func(); last run {_iso_time(last_run)}")
                return (True, time.time())
        obj.logger_for().debug(f"Task {self.description} last run for {obj} at {_iso_time(last_run)}; re-running not required")
        return (False, last_run)

    def invalidator(self, slow=False):
        '''Decorator to indicate  an invalidation function for a :func:`setup_task`

        This decorator indicates a function that will validate whether some setup_task has successfully been created.  As an example, if a setup_task creates a VM, an invalidator could invalidate the task if the VM no longer exists.  Invalidators work as an additional check along side the existing mechanisms to track which setup_tasks are run.  Even if an invalidator  would not invalidate a task, the task would still be performed if its stamp does not exist.  Compare :meth:`check_completed` for a mechanism to exert direct control over whether a task is run.

        :param: slow
            If true, this invalidator is slow to run and should only be run if ``config.expensive_checks`` is True.

        Invalidators should return something True if the task is valid and something falsy to invalidate the task and request that the task and all dependencies be re-run.

        Usage example::

            @setup_task("create_vm)
            async def create_vm(self):
                # ...
            @create_vm.invalidator()
            async def create_vm(self, **kwargs):
                # if VM exists return true else false

        The invalidator receives the following keyword arguments;
        invalidators should be prepared to receive unknown arguments:

        last_run
            The time at which the task was last successfully run


        '''
        def wrap(f):
            self.invalidator_func = f
            return self
        return wrap

    def hash(self):
        '''Provides a mechanism for rerunning setup_tasks when inputs have changed.  Usage::

            @setup_task("do something")
            def do_something(self):
                # do stuff

            @do_something.hash()
            do_something(self):
                # return a rapid hash of the major inputs on which the do_something tasks varies

        Hash functions are entirely ignored when *check_completed* is
        used.  The hash function is called every time Carthage wishes
        to know whether the task has completed, so it needs to be fast
        to compute.  The result of the hash function is stored in the
        completion stamp on successful completion.  On later runs, the
        result of the hash function is checked against the completion
        stamp.  If these two results differ, the task is rerun.
        '''
        def wrap(func):
            self.hash_func = func
            return self
        return wrap

    def check_completed(self):
        '''Decorator to provide function indicating whether a task has already been done

        Usage::

            @setup_task("task")
            async def setup_something(self):
                # do stuff
            @setup_something.check_completed()
            def setup_something(self):
                # Return :func:`time.time` when the task was completed or None or true
                # If True is returned, then task is marked completed, but will not work well with dependencies

        '''
        def wrap(f):
            self.check_completed_func = f
            return self
        return wrap


class TaskWrapper(TaskWrapperBase):

    def __init__(self, func, **kwargs):
        super().__setattr__('func', func)
        super().__init__(**kwargs)

    @memoproperty
    def stamp(self):
        return self.func.__name__

    def __getattr__(self, a):
        if a == "func":
            raise AttributeError
        return getattr(self.func, a)

    extra_attributes = frozenset()

    def __setattr__(self, a, v):
        if a in ('func', 'stamp', 'order',
                 'invalidator_func', 'check_completed_func', 'hash_func') or a in self.__class__.extra_attributes:
            return super().__setattr__(a, v)
        else:
            return setattr(self.func, a, v)

    @property
    def __wraps__(self):
        return self.func


class TaskMethod:

    def __init__(self, task, instance):
        self.task = task
        self.instance = instance

    def __call__(self, *args, **kwargs):
        return self.task(self.instance, *args, **kwargs)

    def __getattr__(self, a):
        return getattr(self.task.func, a)

    def __repr__(self):
        return f"<TaskMethod {self.task.stamp} of {self.instance}>"


def setup_task(description, *,
               order=None,
               before=None):
    '''Mark a method as a setup task.  Describe the task for logging.  Must be in a class that is a subclass of
    SetupTaskMixin.  Usage::

        @setup_task("unpack"
        async def unpack(self): ...

    :param order: Overrides the order in which tasks are run; an integer; lower numbered tasks are run first, higher numbered tasks are run later.  It is recommended that task ordering be a total ordering, but this is not a requirement.  It is an error if both *order* and *before* are set.

    :param before: Run this task before the task referenced in *before*.

    '''
    global _task_order
    if order and before:
        raise TypeError('Order and before cannot both be specified')
    if before:
        order = before.order - 1
    if order and order > _task_order:
        _task_order = order
        _inc_task_order()

    def wrap(fn):
        kws = {}
        if order:
            kws['order'] = order
        t = TaskWrapper(func=fn, description=description, **kws)
        return t
    return wrap


class SkipSetupTask(Exception):
    pass


class SetupTaskMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setup_tasks = sorted(self.class_setup_tasks(),
                                  key=lambda t: t.order)

    def add_setup_task(self, task, **kwargs):
        if isinstance(task, TaskWrapperBase):
            if kwargs:
                raise RuntimeError('kwargs cannot be specified if task is a TaskWrapper')
        else:
            stamp = kwargs.pop('stamp', None)
            task = TaskWrapper(func=task, **kwargs)
            if stamp:
                task.stamp = stamp
        self.setup_tasks.append(task)
        # xxx reorder either here or in run_setup_tasks

    async def run_setup_tasks(self, context=None):
        '''Run the set of collected setup tasks.  If context is provided, it
        is used as an asynchronous context manager that will be entered before the
        first task and eventually exited.  The context is never
        entered if no tasks are run.
        '''
        injector = getattr(self, 'injector', carthage.base_injector)
        ainjector = getattr(self, 'ainjector', None)
        if ainjector is None:
            ainjector = injector(AsyncInjector)
        config = getattr(self, 'config_layout', None)
        if config is None:
            config = injector(ConfigLayout)
        context_entered = False
        dry_run = config.tasks.dry_run
        dependency_last_run = 0.0
        for t in self.setup_tasks:
            should_run, dependency_last_run = await t.should_run_task(self, dependency_last_run, ainjector=ainjector)
            if should_run:
                self.injector.emit_event(
                    InjectionKey(SetupTaskMixin), "task_should_run",
                    self, task=t,
                    adl_keys=self.setup_task_event_keys())
                try:
                    if (not context_entered) and context is not None:
                        await context.__aenter__()
                        context_entered = True
                    if not dry_run:
                        self.logger_for().info(f"Running {t.description} task for {self}")
                        with SetupTaskContext(self, t):
                            await ainjector(t, self)
                        dependency_last_run = time.time()
                        self.logger_for().info(f"Finished running {t.description} task for {self} at {dependency_last_run}")
                    else:
                        self.logger_for().info(f'Would run {t.description} task for {self}')
                except SkipSetupTask:
                    pass
                except Exception:
                    self.logger_for().exception(f"Error running {t.description} for {self}:")
                    if context_entered:
                        await context.__aexit__(*sys.exc_info())
                    raise
            else:           # not should_run
                self.injector.emit_event(
                        InjectionKey(SetupTaskMixin), "task_already_run",
                        self, task=t,
                        adl_keys=self.setup_task_event_keys())
        if context_entered:
            await context.__aexit__(None, None, None)

    def setup_task_event_keys(self):
        '''Yield the set of keys that setup_task related events should be dispatc should be dispatched to.  In addition to keys yi.yielded by this generator, all setup tasks events are dispatched to InjectionKey(SetupTaskMixin).
        '''
        yield self.default_class_injection_key()
        yield from self.supplementary_injection_keys(self.default_class_injection_key())
        
        
    @classmethod
    def class_setup_tasks(cls):
        '''
        returns a tuple of TaskWrapperBases associated with *cls*.  Handles MRO and methods overriding tasks.
        This is a relatively expensive operation, so it is memoized per class.
        '''
        try:  return cls.__dict__['_class_setup_tasks_prop']
        except (AttributeError, KeyError): pass
        results: list[TaskWrapperBase] = []
        meth_names = {}
        for c in cls.__mro__:
            if not issubclass(c, SetupTaskMixin):
                continue
            for m in c.__dict__:
                if m in meth_names:
                    continue
                meth = getattr(c, m)
                meth_names[m] = True
                if isinstance(meth, TaskWrapperBase):
                    results.append(meth)

        cls._class_setup_tasks_prop = tuple(results)
        return cls.__dict__['_class_setup_tasks_prop']
    
    async def async_ready(self):
        '''
        This may need to be overridden, but is provided as a default
        '''
        await self.run_setup_tasks()
        return await super().async_ready()

    def create_stamp(self, stamp, contents):
        try:
            with open(os.path.join(self.stamp_path, ".stamp-" + stamp), "wt") as f:
                # on NFS, opening a zero-length file even for truncate does not reset the utime
                os.utime(f.fileno())
                if contents:
                    f.write(contents)
        except FileNotFoundError:
            os.makedirs(self.stamp_path, exist_ok=True)
            with open(os.path.join(self.stamp_path, ".stamp-" + stamp), "wt") as f:
                os.utime(f.fileno())
                if contents:
                    f.write(contents)

    def delete_stamp(self, stamp):
        try:
            os.unlink(os.path.join(self.stamp_path, ".stamp-" + stamp))
        except FileNotFoundError:
            pass

    def check_stamp(self, stamp, raise_on_error=False):
        '''
        :returns: a tuple containing the unix time of the stamp and the tex t contents of the stamp.  The first element of the tuple is False if the stamp does not exist
        '''
        if raise_on_error not in (True, False):
            raise SyntaxError(f'raise_on_error must be a boolean. current value: {raise_on_error}')
        try:
            path = Path(self.stamp_path) / f'.stamp-{stamp}'
            res = os.stat(path)
        except FileNotFoundError:
            if raise_on_error:
                raise RuntimeError(f"stamp directory '{self.stamp_path}' did not exist") from None
            return (False, "")
        return res.st_mtime, path.read_text()

    def logger_for(self):
        try:
            return self.logger
        except AttributeError:
            return logger

    def inspect_setup_tasks(self, *, stamp_stem="", instance_id=None):
        '''Iterates over the setup tasks of a ninstance and provides an inspector that can determine if a task would run and what its description is.
        '''
        prev = None
        for t in self.setup_tasks:
            prev = TaskInspector(task=t, from_obj=self, previous=prev)
            if stamp_stem: prev.stamp = stamp_stem+prev.stamp
            if instance_id: prev.instance_id = instance_id
            yield prev

    @classmethod
    def inspect_class_setup_tasks(cls):
        '''Iterates over all the setup tasks that instances of this class will have and provides an inspector for that task.
        '''
        prev = None
        for t in cls.class_setup_tasks():
            prev=TaskInspector(task=t, from_obj=cls, previous=prev)
            yield prev
            

def _iso_time(t):
    return datetime.datetime.fromtimestamp(t).isoformat()


class cross_object_dependency(TaskWrapper):

    '''
    Usage::

        # in a client machine's class
        fileserver_dependency = cross_object_dependency(FileServer.update_files, 'fileserver')

    :param task: a :class:`TaskWrapper`, typically associated with another class, or alternatively the name of an attribute on *relationship* containing either a :class:`TaskWrapper` or :class:`TaskMethod`.

    :param relationship: The string name of a relationship such that calling the *relationship* method on an instance containing this dependency will yield the instance containing *task* that we want to depend on.

    '''

    dependent_task: typing.Union[TaskWrapper, TaskMethod, str]
    relationship: str

    def __init__(self, task, relationship, **kwargs):
        if isinstance(task,str):
            description = task
        else: description=task.description
        super().__init__(func=lambda self: None,
                         description=f'Dependency on `{description}\' task of {relationship}',
                         **kwargs)
        self.dependent_task = task
        self.relationship = relationship

    def _handle_task(self, instance):
        if isinstance(self.dependent_task, str):
            relationship = getattr(instance, self.relationship)
            self.dependent_task = getattr(relationship, self.dependent_task)
            if isinstance(self.dependent_task, TaskMethod):
                self.dependent_task = self.dependent_task.task
                
    @inject(ainjector=AsyncInjector)
    async def check_completed_func(self, instance, ainjector):
        self._handle_task(instance)
        task = self.dependent_task
        should_run, last_run = await task.should_run_task(getattr(instance, self.relationship), ainjector=ainjector)
        # We don't care about whether the task would run again, only when it last run.
        if last_run > 0.0:
            return last_run
        # We have no last_run info so we don't know that we need to trigger a re-run
        return True

    def __repr__(self):
        if isinstance(self.dependent_task, str):
            description = self.dependent_task
        else: description = self.dependent_task.description
        return f'<Depend on {description} task of {self.relationship}>'


class mako_task(TaskWrapper):

    '''
    Usage::

        dnsmasq_task = mako_task('dnsmasq.conf.mako',
            network = InjectionKey("some_network"))

    Typically used in a :class:`~carthage.modeling.MachineModel`.  Introduces a setup task to render a mako template.  Extra keyword arguments can be :class:`InjectionKey` in which case they are instantiated in the context of the injector of the object to which the setup task is attached.  These arguments are made available in the mako template context.  The *instance* template context argument is introduced and points to  the object on which the setup task is run.

If the template has a def called *hash*, this def will be rendered with the same arguments as the main template body.  This value will be stored in the completion stamp; if the hash changes, the template will be re-rendered.  For performance reasons, try to keep the hash easy to compute.

    '''

    template: str
    output: str

    extra_attributes = frozenset({'template', 'output',
                                  })

    def __init__(self, template, output=None, **injections):
        kwargs = {}
        # Split kwargs; Leading _ is left as arguments to setup_task,
        # others are injections.
        for k in injections:
            if k.startswith("_"):
                kwargs[k[1:]] = injections.pop(k)

        # A separate function so that injection works; consider
        # TaskMethod.__setattr__ to understand.
        @inject(**injections)
        def func(*args, **kwargs):
            return self.render(*args, **kwargs)

        @inject(**injections)
        def hash_func(instance, **kwargs):
            template = self.lookup.get_template(self.template)
            if template.has_def('hash'):
                hash_template = template.get_def("hash")
                return hash_template.render(instance=instance, **kwargs)
            else:
                return template.render(instance=instance, **kwargs)
        self.template = template
        if output is None:
            output = template
            if output.endswith('.mako'):
                output = output[:-5]
        self.output = output
        super().__init__(func=func,
                         description=f'Render {self.template} template',
                         hash_func=hash_func,
                         **kwargs)

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        import sys
        import mako.lookup
        module = sys.modules[owner.__module__]
        try:
            self.lookup = module._mako_lookup
        except AttributeError:
            if hasattr(module, '__path__'):
                resources = import_resources_files(module)
            elif module.__package__ == "":
                resources = Path(module.__file__).parent
            else:
                resources = import_resources_files(module.__package__)
            templates = resources / 'templates'
            if not templates.exists():
                templates = resources
            module._mako_lookup = mako.lookup.TemplateLookup([str(templates)], strict_undefined=True)
            self.lookup = module._mako_lookup

    def render(task, instance, **kwargs):
        template = task.lookup.get_template(task.template)
        output = Path(instance.stamp_path).joinpath(task.output)
        os.makedirs(output.parent, exist_ok=True)
        with open(output, "wt") as f:
            f.write(template.render(
                instance=instance,
                **kwargs))


def find_mako_tasks(tasks):
    for t in tasks:
        if isinstance(t, mako_task):
            yield t


def install_mako_task(relationship, cross_dependency=True, *, relationship_ready=True):
    '''
:param relationship: The name of an attribute property containing :class:`mako_tasks <mako_task>` in its :meth:`~SetupTaskMixin.setup_tasks`.

    :param cross_dependency: If true (the default), rerun the installation whenever any of the underlying mako_tasks change.

    :param relationship_ready:  Bring the relationship to *async_become_ready* state before  installing mako tasks or checking cross-object dependencies.  Bringing the relationship to ready state allows the hash block in the mako template to access parts of the instance that are set up as part of bringing the instante to ready.  It also guarantees that the mako templates will be generated at least once.  However, it can produce dependency loops that hange Carthage in some circumstances.
    
    This task is generally associated on a machine to install mako templates rendered on the model.  Typical usage might look like::

        install_mako = install_mako_task('model')

    '''
    if relationship == 'self' and relationship_ready:
        # We could just set relationship_ready to false but if we do that in thde easy case of self, we will miss harder cases like host.
        raise RuntimeError('Set relationship_ready=False when using self as a relationship')
    
    @setup_task("Install mako templates")
    async def install(self):
        async with self.filesystem_access() as fspath:
            if relationship == 'self':
                related = self
            else:
                related = getattr(self, relationship)
            if relationship_ready: await related.async_become_ready()
            base = Path(related.stamp_path)
            path = Path(fspath)
            for mt in find_mako_tasks(related.setup_tasks):
                if os.path.isabs(mt.output):
                    logger.warn(f'{mt} has absolute path; skipping install')
                    continue
                src = base / mt.output
                dest = path / mt.output
                os.makedirs(dest.parent, exist_ok=True)
                shutil.copy2(src, dest)
    if cross_dependency:
        @install.invalidator()
        @inject(ainjector=AsyncInjector)
        async def install(self, ainjector, last_run, **kwargs):
            if relationship == 'self':
                related = self
            else:
                related = getattr(self, relationship)
            if relationship_ready: await related.async_become_ready()
            last = 0.0
            for mt in find_mako_tasks(related.setup_tasks):
                run, last = await mt.should_run_task(related, dependency_last_run=last, ainjector=ainjector)
                if run:
                    return False
                if last > last_run:
                    return False
            return True
    return install



# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import pytest
from carthage.pytest import *
from carthage.dependency_injection import *
from carthage.plugins import load_plugin
from pathlib import Path


dir = Path(__file__).parent


@pytest.fixture()
def injector(ainjector):
    return ainjector.injector


def test_load_vmware(injector):
    injector(load_plugin, "carthage.vmware")


def test_dir_plugin(injector):
    injector(load_plugin, dir / "test_plugin")



# Copyright (C) 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import argparse
from pathlib import Path
from .utils import carthage_main_argparser, mako_lookup

SKELETONS = ('layout',
             )

__all__ = []

def skeleton_subparser_setup(parser: argparse.ArgumentParser) ->argparse.ArgumentParser:
    action = parser.add_subparsers(title='layout', dest='skel',
                                   help='which layout to select', 
                                   required=True)
    for skel in SKELETONS:
        plugin_uri = mako_lookup.adjust_uri(skel+'/carthage_plugin.yml', 'skeletons/')
        plugin_template = mako_lookup.get_template(plugin_uri+'.mako')
        help_text = plugin_template.get_def('help').render()
        subparser = action.add_parser(skel,
                                         help=help_text)
        if plugin_template.has_def('arguments'):
            arguments = plugin_template.get_def('arguments')
            # allow template to adjust arguments.
            arguments.render(parser=subparser)
    return parser

__all__ += ['skeleton_subparser_setup']

def subrender_wrapper(context, relative_in, relative_out:Path, absolute_out:Path):
    def subrender(template, output=None, **kwargs):
        new_ctx = context.copy()
        new_ctx.update(kwargs)
        del new_ctx['subrender']
        template_uri = mako_lookup.adjust_uri(template, relative_in)
        our_template = mako_lookup.get_template(template_uri+'.mako')
        if output is None and our_template.has_def('output'):
            output_text = our_template.get_def('output').render(**new_ctx)
            output = output_text.strip()
        if output is None:
            assert not template.startswith('/')
            output = relative_out.joinpath(template)
        else:
            output = Path(output)
            if output.is_absolute(): # map / to  absolute_out
                output = output.relative_to('/')
                output = absolute_out.joinpath(output)
            else:
                output = relative_out.joinpath(output)
        output.parent.mkdir(parents=True, exist_ok=True)
        new_ctx['subrender'] = subrender_wrapper(new_ctx, our_template.uri,
                                                 output.parent, absolute_out)
        output.write_text(our_template.render(**new_ctx))
        return ""
    return subrender

def render_skeleton(skeleton, output_top, args, **kwargs):
    context = kwargs
    context['args'] = args
    assert skeleton in SKELETONS
    context['subrender'] = None
    subrender = subrender_wrapper(context,
                                  'skeletons/'+skeleton+'/',
                                  Path(output_top), Path(output_top))
    subrender('carthage_plugin.yml')
    
    
__all__ += ['render_skeleton']

def package_to_dir(p):
    return p.replace('.','/')




# Copyright (C) 2018, 2019, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import weakref
from .dependency_injection import inject, Injectable
from .config import ConfigLayout
from .container import ssh_origin
from . import sh


@inject(config_layout=ConfigLayout)
class PortReservation:

    _ports_used = weakref.WeakSet()
    __slots__ = ('port', '__weakref__')

    def __init__(self, config_layout):
        super().__init__()
        found = False
        for i in range(config_layout.min_port, config_layout.min_port + config_layout.num_ports + 1):
            self.port = i
            if self not in self._ports_used:
                self._ports_used.add(self)
                found = True
                break
        if not found:
            raise NoPortsError("No free ports")

    def __hash__(self):
        return hash(self.port)

    def __eq__(self, other):
        # Assumes that within the scope we care about for equality, a
        # given port reservation is unique.  If that's not true for a
        # subclass it's a good hint that containment rather than
        # subclassing is the right approach.
        if isinstance(other, int):
            return other == self.port
        elif isinstance(other, PortReservation):
            return self.port == other.port
        else:
            return super().__eq__(other)


@inject(config_layout=ConfigLayout,
        ssh_origin=ssh_origin)
class ExposedPort(PortReservation):

    __slots__ = ('expose_process',)

    bind_addr = '127.0.0.1'

    def __init__(self, dest_addr, *, config_layout, ssh_origin):
        super().__init__(config_layout=config_layout)
        self.expose_process = sh.socat(
            "tcp-listen:{},bind={bind},fork,reuseaddr".format(self.port, bind=self.bind_addr),
            'exec:"nsenter -t {leader} -n -i -m socat stdio {dest}",nofork'.format(
                leader=ssh_origin.container_leader,
                dest=dest_addr.replace('"', "\\\"")),
            _bg=True, _bg_exc=False)

    def close(self):
        if self.expose_process is not None:
            self.expose_process.kill()
            self.expose_process = None

    def __del__(self):
        self.close()



# Copyright (C) 2018, 2019, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import time
import pyVim.task


def wait_for_task(task):
    loop = asyncio.get_event_loop()
    ''' Returns a future that  when done indicates the task is complete.  Note that while this is not async, it should be treated as if it is because it returns a future.
    Example usage::

        await wait_for_task(task)

    '''
    # We use a separate thread to avoid blocking the async loop on http round trips to look up task state
    def callback():
        pyVim.task.WaitForTask(task)
        if task.info.state == 'error':
            class TaskError(type(task.info.error)):
                def __str__(self):
                    return f'Error: {super().__str__()}; info: {self.info_str}'

                def __init__(self, task):
                    self.__dict__.update(task.info.error.__dict__)
                    self.__dict__['info_str'] = str(task.info)
                    self.__dict__['task'] = task
                    task.info.__dict__['error'] = None
            raise TaskError(task)
    return loop.run_in_executor(None, callback)


class TaskIgnoreErrors:

    def __init__(self, *types):
        self.types = types

    def __enter__(self):
        print('enter types', self.types)
        return self

    def __exit__(self, etype, einstance, etraceback):
        if etype is not None:
            for t in self.types:
                if issubclass(etype, t):
                    return True



# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from pathlib import Path
import os
import shutil
from carthage.dependency_injection import *
from carthage.machine import AbstractMachineModel
from carthage.setup_tasks import *
from .network import NetworkLink, hash_network_links
from .utils import mako_lookup
from carthage import ConfigLayout, sh
import logging

logger = logging.getLogger('carthage.network')

local_type_map = dict(
    bond=dict(
        netdev="bond-netdev.mako",
        member_network="bond-network.mako"),
    bridge=dict(
        netdev="bridge-netdev.mako",
        member_network="bridge-network.mako",
    ),
    xfrm=dict(
        netdev='xfrm-netdev.mako',
    ),
    physical=dict(
        link="physical-link.mako",
    ),
    vlan=dict(
        netdev="vlan-netdev.mako",
        member_network="vlan-network.mako"),
    none=dict(network=None)
)


def templates_for_link(l: NetworkLink):
    templates = dict(
        network="default-network.mako",
    )
    templates.update(local_type_map.get(l.local_type or 'physical', {}))
    for ml in l.member_of_links:
        if ml.local_type:
            map_entry = local_type_map.get(ml.local_type, {})
            for i in ('network',):
                if 'member_' + i in map_entry:
                    templates[i] = map_entry['member_' + i]
    return templates


class NotNeeded(Exception):
    '''
    Indicates that a particular template is not needed and should be skipped.  Raised in the template itself.
    '''
    pass


@inject(config_layout=ConfigLayout)
class SystemdNetworkModelMixin(SetupTaskMixin, AsyncInjectable):

    def __init__(self, **kwargs):
        from .machine import AbstractMachineModel
        super().__init__(**kwargs)
        if isinstance(self, AbstractMachineModel) and hasattr(self, 'machine_mixins'):
            self.machine_mixins = list(self.machine_mixins) + [SystemdNetworkInstallMixin]

    @setup_task("Generate Network Configuration")
    async def generate_network_config(self):
        await self.resolve_networking()
        networking_dir = Path(self.stamp_path) / "networking"
        try:
            shutil.rmtree(networking_dir)
        except FileNotFoundError:
            pass
        os.makedirs(networking_dir)
        for link in self.network_links.values():
            self._render_network_configuration(link, networking_dir)

    @generate_network_config.hash()
    def generate_network_config(self):
        return "20221106" + str(hash_network_links(self.network_links))

    def _render_network_configuration(self, link: NetworkLink, dir: Path):
        templates = templates_for_link(link)
        for ext, template_name in templates.items():
            if ext.startswith('member_'):
                continue
            if template_name is None:
                continue
            template = mako_lookup.get_template(template_name)
            try:
                rendering = template.render(
                    link=link,
                    logger=logger,
                    NotNeeded=NotNeeded
                )
                output_fn = dir.joinpath(f"10-carthage-{link.interface}.{ext}")
                with output_fn.open("wt") as f:
                    f.write(rendering)
            except NotNeeded:
                pass


class SystemdNetworkInstallMixin(SetupTaskMixin):

    generate_config_dependency = cross_object_dependency(SystemdNetworkModelMixin.generate_network_config, 'model')

    @setup_task("Install Systemd Networking")
    async def install_systemd_networking(self):
        async with self.filesystem_access() as path:
            try:
                networking_dir = Path(self.model.stamp_path) / "networking"
            except AttributeError:
                networking_dir = Path(self.stamp_dir) / "networking"
            if not networking_dir.exists():
                raise SkipSetupTask
            await sh.rsync(
                "-a", "--delete",
                "--filter=include 10-carthage-*",
                "--filter=exclude,! 10-carthage-*",
                str(networking_dir) + "/",
                Path(path) / "etc/systemd/network",
                _bg=True,
                _bg_exc=False)


__all__ = [
    'SystemdNetworkModelMixin',
    'SystemdNetworkInstallMixin',
]



# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import dataclasses
import typing
import weakref
from .base import NetworkLink
from ..dependency_injection import *


class BondLink(NetworkLink):

    local_type = "bond"


class BridgeLink(NetworkLink):

    local_type = "bridge"

    vlan_filter: typing.Optional[bool] = False


@dataclasses.dataclass(eq=False)
class VlanLink(NetworkLink):

    local_type = "vlan"
    vlan_id: typing.Optional[int] = None

    @classmethod
    def validate_subclass(cls, args, unresolved):
        if not unresolved:
            if 'vlan_id' not in args and (not args['net'].vlan_id):
                raise ValueError("vlan_id must be specified directly or on the network")
        return super().validate_subclass(args, unresolved)

    def __init__(self, connection, interface, args):
        if 'vlan_id' not in args:
            args['vlan_id'] = args['net'].vlan_id
        super().__init__(connection, interface, args)


class NoLocalLink(NetworkLink):
    local_type = 'none'


@dataclasses.dataclass(init=False, eq=False)
class XfrmLocalLink(NetworkLink):

    '''A link corresponding to a linux xfrm interface'''
    interface_id: typing.Optional[int] = None
    local_type = 'xfrm'



# Copyright (C) 2018, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage.utils import memoproperty


def test_memo_prop():
    class m:
        @memoproperty
        def foo(self):
            nonlocal called
            assert called is False
            called = True
            return 99
    called = False
    assert isinstance(m.foo, memoproperty)
    mo = m()
    assert mo.foo == 99
    assert mo.foo == 99  # and not called a second time



# Copyright (C) 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import importlib
import logging
import types
import typing
import sys
import yaml
from pathlib import Path
from importlib.util import spec_from_file_location, module_from_spec, find_spec
from urllib.parse import urlparse
from .dependency_injection import *
from .config import ConfigLayout
from .files import checkout_git_repo


logger = logging.getLogger('carthage.plugins')


class CarthagePlugin(Injectable):

    name: str
    package: typing.Optional[importlib.resources.Package]
    resource_dir: Path
    metadata: dict
    import_error: str = None

    def __init__(self, name: str, package: importlib.resources.Package,
                 metadata: dict,
                 **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.package = package
        if 'resource_dir' in metadata:
            self.resource_dir = Path(metadata['resource_dir'])
        else:
            self.resource_dir = Path(package.__path__[0])
        self._resources = {}
        self.metadata = metadata

    def _get_resource(self, resource):
        p = self.resource_dir.joinpath(resource)
        try:
            return self._resources[p]
        except BaseException:
            if p.exists():
                self._resources[p] = p
                return p
            else:
                self._resources[p] = None
                return None

    def contains_resource(self, resource):
        return bool(self._get_resource(resource))

    # Yes this fails for zips and similar.
    # If we care, we'd need to go to a lot of trouble to unpack things,
    # because we do need to make directories available for things like ansible
    # plays.
    def resource_path(self, resource):
        return self._get_resource(resource)


@inject(injector=Injector)
def load_plugin(spec: str,
                *, injector,
                ignore_import_errors=False):
    '''
    Load a plugin from a plugin specification:

    * A path name to a directory containing a ``carthage_plugin.yml``

    * A python package name

    * A ``https`` URL to a Git repository

    * A ``git+ssh`` URL to a git repository

    :param ignore_import_errors:  If True, succeed and register the plugin even if the python code raises.  This is intended to allow the plugin to be loaded so its metadata can be examined to determine dependencies.  Obviously the plugin is unlikely to be functional in such a state.
    '''
    if ':' in str(spec):
        spec = handle_plugin_url(str(spec), injector)
    if hasattr(spec, "__fspath__") or '/' in spec or spec == '.' or spec == '..':
        path = Path(spec).resolve()
        metadata_path = path / "carthage_plugin.yml"
        if not metadata_path.exists():
            raise FileNotFoundError(f'{metadata_path} not found')
        metadata = yaml.safe_load(metadata_path.read_text())
        if 'resource_dir' not in metadata:
            metadata['resource_dir'] = path
        if 'name' not in metadata:
            raise ValueError(f'metadata must contain a name when loading plugin from path')
        # Stop early if already loaded
        try:
            injector.get_instance(InjectionKey(CarthagePlugin, name=metadata['name']))
            logger.debug(f'Plugin {metadata["name"]} already loaded')
            return
        except KeyError:
            pass
        _handle_plugin_config(injector, metadata, metadata_path, ignore_import_errors=ignore_import_errors)
        try:
            python_path = metadata['python']
            python_path = str(path.joinpath(python_path))
            if python_path not in sys.path:
                sys.path.insert(0, python_path)
        except KeyError:
            pass
        if 'package' in metadata:
            module_spec = find_spec(metadata['package'])
        else:
            package_path = path.joinpath("carthage_plugin.py")
            name = metadata['name']
            if '.' not in name:
                name = "carthage.carthage_plugins." + name
                _setup_carthage_plugins_module()
            if package_path.exists():
                module_spec = spec_from_file_location(
                    name, location=package_path,
                    submodule_search_locations=[str(path / "python")]
                )
            else:
                module_spec = None
    else:  # spec is a package
        module_spec = find_spec(spec)
        metadata = None
    package = None
    import_error = None
    if module_spec:
        # For some reason module_from_spec sometimes changes spec.name
        module_name = module_spec.name
        if module_name in sys.modules:
            package = sys.modules[module_name]
        else:
            package = module_from_spec(module_spec)
            try:
                parent_module = None
                parent, _, stem = module_name.rpartition('.')
                if parent:
                    parent_module = importlib.import_module(parent)
                    setattr(parent_module, stem, package)
                    
                sys.modules[module_name] = package
                module_spec.loader.exec_module(package)
            except BaseException as e:
                try: del sys.modules[module_name]
                except KeyError: pass
                if parent_module: delattr(parent_module, stem)
                if ignore_import_errors:
                    logger.debug('Ignoring error importing %s: %s', module_spec.name, str(e))
                    import_error = str(e)
                else:
                    raise
    return injector(load_plugin_from_package, package, metadata,
                    ignore_import_errors=ignore_import_errors,
                    import_error=import_error)


def handle_plugin_url(url, injector):
    parsed = urlparse(url)
    if parsed.scheme in ('https', 'git+ssh'):
        return handle_git_url(parsed, injector)
    else:
        raise NotImplementedError(f"Don't know how to handle {parsed.scheme} URL")


def handle_git_url(parsed, injector):
    config = injector(ConfigLayout)
    stem = Path(parsed.path).name
    if stem.endswith('.git'):
        stem = stem[:-4]
    dest = Path(config.checkout_dir) / stem
    if dest.exists():
        return dest
    logger.info(f'Checking out {parsed.geturl()}')
    injector(checkout_git_repo, parsed.geturl(), dest, foreground=True)
    return dest


@inject(injector=Injector)
def load_plugin_from_package(package: typing.Optional[types.ModuleTyp],
                             metadata: dict = None,
                             *, ignore_import_errors=False,
                             import_error=None, 
                             injector):
    if (not metadata) and (not package):
        raise RuntimeError('Either package or metadata must be supplied')
    if metadata:
        if 'resource_dir' in metadata:
            metadata_path = Path(metadata['resource_dir']) / "carthage_plugin.yml"
        else:
            metadata_path = Path(package.__file__)
    if not metadata:
        if not package.__spec__.origin:
            raise SyntaxError(f'{package.__name__} is not a Carthage plugin')
        try:
            metadata = yaml.safe_load(importlib.resources.read_text(
                package, "carthage_plugin.yml"))
            metadata_path = package.__file__
        except (FileNotFoundError, ImportError):
            # consider the case of hadron-operations
            # plugin is hadron.carthage
            # but when not installed resources live at the top level of the checkout.
            components = len(package.__name__.split("."))
            path_root = Path(package.__file__).parents[components]
            if path_root.joinpath("carthage_plugin.yml").exists():
                metadata = yaml.safe_load(path_root.joinpath("carthage_plugin.yml").read_text())
                metadata_path = path_root.joinpath('carthage_plugin.yml')
                if 'resource_dir' not in metadata:
                    metadata['resource_dir'] = path_root
            else:
                metadata = {}
                metadata_path = None

    if 'name' in metadata:
            name = metadata['name']
    else:
        name = package.__name__
    try:
        injector.get_instance(InjectionKey(CarthagePlugin, name=name))
        return # already loaded
    except KeyError: pass
    _handle_plugin_config(injector=injector, metadata=metadata, path=metadata_path, ignore_import_errors=ignore_import_errors)
    try:
        plugin_module = importlib.import_module(".carthage_plugin", package=package.__name__)
    except (ImportError, AttributeError):
        plugin_module = package
        # note plugin_module may be none if package is none
    plugin_func = getattr(plugin_module, "carthage_plugin", None)
    if not any((plugin_func, metadata)):
        raise SyntaxError(f'{package.__file__} is not a Carthage plugin')

    if package and 'package' not in metadata:
        metadata['package'] = package.__name__
    if plugin_func:
        try:
            res = injector(plugin_func)
        except Exception as e:
            res = None
            if not ignore_import_errors: raise
            if not import_error: import_error = e
    else:
        res = None
    if isinstance(res, CarthagePlugin):
        assert res.name == name, "Metadata name must agree with resulting plugin for duplicate load detection to work"
        plugin_object = res
    else:
        plugin_object = injector(CarthagePlugin, name=name, package=package, metadata=metadata)
    if import_error: plugin_object.import_error = import_error
    injector.add_provider(
        InjectionKey(CarthagePlugin, name=plugin_object.name),
        plugin_object)


def _handle_plugin_config(injector, metadata, path, ignore_import_errors):
    # we don't want to take a ConfigLayout as a dependency because
    # that tends to push its instantiation too high in the injector
    # hierarchy
    config = injector(ConfigLayout)
    if 'config' in metadata:
        config.load_yaml(yaml.dump(metadata['config']), path=path, ignore_import_errors=ignore_import_errors)


def _setup_carthage_plugins_module():
    from types import ModuleType
    sys.modules['carthage.carthage_plugins'] = ModuleType('carthage.carthage_plugins')
__all__ = ['load_plugin', 'load_plugin_from_spec']



# Copyright (C) 2019, 2020, 2021, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import importlib
import yaml
from pathlib import Path
from ..dependency_injection import inject, Injectable, InjectionKey, Injector, partial_with_dependencies


from .schema import config_key, ConfigAccessor, ConfigSchema


@inject(injector=Injector)
class ConfigLayout(ConfigAccessor, Injectable):

    def __init__(self, injector):
        super().__init__(injector, "")

    def _load(self, d, injector, into, prefix):
        for k, v in d.items():
            full_key = prefix + k
            if full_key in ConfigSchema._schemas:
                if not isinstance(v, dict):
                    raise ValueError("{} should be a dictionary".format(full_key))
                self._load(v, injector, ConfigSchema._schemas[prefix + k], prefix + k + ".")
            else:
                try:
                    schema_item = into[k]
                    # We cannot simply use the yaml provided value
                    # because the type class in the schema item may
                    # process the value in construction.  For example
                    # ConfigString does substitutions in construction
                    # but ultimately returns a real string.  We cannot
                    # simply construct the value here, because there
                    # may be forward reference substitutions.  So we
                    # need to defer actually constructing the value
                    # until it is used.  We do this by creating an
                    # injectible class that acts as a dependency
                    # provider for the configuration key.

                    class ConfigValue(schema_item.type, Injectable):
                        new_value = v

                        def __new__(self, **kwargs):
                            return super().__new__(self, self.new_value, **kwargs)

                        def __init__(self, **kwargs):
                            # This is a bit complicated because of the
                            # various cases.  It appears that if a
                            # class never overrides __init__ at all,
                            # type.__new__ may not end up calling
                            # __init__.  But if any args or kwargs
                            # chain back to object.__init__, then
                            # construction will raise.  Unfortunately,
                            # Injectable needs an __init__.  So if the
                            # underlying type is something like str,
                            # without __init__, then we need to not
                            # pass args along to super().__init__.  If the underlying type is list or
                            # something with __init__ provided by C code, we need to pass along a
                            # value, but the only way we can detect this is if __init__ has no
                            # __function__ attribute.  If the underlying type is something that does
                            # have an __init__ written in python we need  to pass along args, but in
                            # that case, we detect a __function__ attribute that is not
                            # Injectable.__init__.
                            sup_obj = super()
                            if getattr(sup_obj.__init__, '__func__', None) == Injectable.__init__:
                                # Nothing will eat the value, so it will pass along to Object.__init__ and fail
                                sup_obj.__init__(**kwargs)
                            else:
                                sup_obj.__init__(self.new_value, **kwargs)

                    injector.replace_provider(config_key(full_key), ConfigValue)
                except AttributeError:
                    raise AttributeError("{} is not a config attribute".format(full_key)) from None

    def load_yaml(self, y, *, injector=None, path=None,
                  ignore_import_errors=False):
        '''
        :param ignore_import_errors: If true, then loading a plugin will not fail simply because the plugin's python code  raises an error.  This is intended to allow introspection of plugin metadata to determine plugin dependencies; actually trying to use a plugin that has raised an error on load is unlikely to work.
        '''
        if injector is None:
            injector = self._injector
        if path:
            base_path = Path(path).parent
        else:
            base_path = Path(y.name).parent
        d = yaml.safe_load(y)
        assert isinstance(d, dict)
        from .types import ConfigPath
        # The plugin loader needs checkout_dir, but we need to
        # load plugins before loading config because plugins can
        # introduce new schema.  This is not strictly correct
        # because the loaded value for checkout_dir may include
        # substitutions to other items that are also in the
        # config.  Don't do that.
        for early_key in ('checkout_dir', 'base_dir'):
            if early_key not in d:
                continue
            setattr(self, early_key,
                    injector(ConfigPath, d[early_key]))
        if 'plugins' in d:
            for p in d['plugins']:
                if (not ':' in p) and (p == '..' or p == '.' or '/' in p):
                    p = base_path.joinpath(p)
                enable_plugin(p, ignore_import_errors=ignore_import_errors)
            del d['plugins']
        if 'include' in d:
            for include in d['include']:
                include = base_path.joinpath(include)
                with include.open("rt") as include_file:
                    self.load_yaml(include_file, ignore_import_errors=ignore_import_errors)
            del d['include']
        try: self._load(d, injector, self._schema, "")
        except KeyError:
            if not ignore_import_errors: raise
            # We can fail to get new schema defined because of an
            # import that failed, so also ignore KeyError from _load.


def enable_plugin(plugin, ignore_import_errors=False):
    '''
Load and enable a Carthage plugin.

    :param plugin: String representing a module containing a :func:`carthage_plugin` entry point.

'''
    from .. import base_injector
    from ..plugins import load_plugin
    base_injector(load_plugin, plugin, ignore_import_errors=ignore_import_errors)
    from . import inject_config
    inject_config(base_injector)


__all__ = ('ConfigLayout',)



# Copyright (C) 2018, 2019, 2020, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import contextlib
import logging
import os
import re
import shutil
import sys
from pathlib import Path
from .dependency_injection import *
from .image import SetupTaskMixin, setup_task, SkipSetupTask, ContainerVolume
from . import sh, ConfigLayout
from .utils import memoproperty
from .machine import MachineRunning, Machine, SshMixin, ssh_origin
import carthage.network
import carthage.ssh

_resources_path = Path(__file__).parent.joinpath("resources")
logger = logging.getLogger('carthage.container')


container_image = InjectionKey('container-image')
container_volume = InjectionKey('container-volume')


@inject(image=InjectionKey(container_image, _ready=False),
        loop=asyncio.AbstractEventLoop,
        config_layout=ConfigLayout,
        network_config=InjectionKey(carthage.network.NetworkConfig, optional=True),
        injector=Injector)
class Container(Machine, SetupTaskMixin):

    def __init__(self, name, *, network_config,
                 skip_ssh_keygen=False, **kwargs):
        super().__init__(
            name=name, **kwargs)
        self.process = None
        self.skip_ssh_keygen = skip_ssh_keygen
        self.running = False
        self._operation_lock = asyncio.Lock()
        self._out_selectors = []
        self._done_waiters = []
        self.container_running = self.machine_running
        self.network_namespace = None
        self.close_volume = True
        self.cleanup_future = None

    rsync_uses_filesystem_access = True

    async def async_ready(self):
        try:
            vol = await self.ainjector.get_instance_async(container_volume)
            self.close_volume = False
        except KeyError:
            await self.image.async_become_ready()
            vol = await self.ainjector(ContainerVolume,
                                       clone_from=self.image,
                                       name="containers/" + self.name)
            self.injector.add_provider(container_volume, vol)
        self.volume = vol
        await self.resolve_networking()
        await self.is_machine_running()
        await self.run_setup_tasks()
        return await super().async_ready()

    async def is_machine_running(self):
        try:
            self.container_leader
            self.running = True
        except sh.ErrorReturnCode_1:
            self.running = False
        return self.running

    @memoproperty
    def stamp_path(self):
        if self.volume is None:
            raise RuntimeError('Volume not yet created')
        return self.volume.path

    async def do_network_config(self, networking):
        if networking and self.network_links:
            namespace = carthage.network.NetworkNamespace(self.full_name, self.network_links)
            try:
                await namespace.start_networking()
                self.network_namespace = namespace
                return ["--network-namespace-path=/run/netns/" + namespace.name,
                        "--capability=CAP_NET_ADMIN",
                        "--resolv-conf=off"]
            except BaseException:
                namespace.close()
                raise
        else:
            return ["--resolv-conf=" + self.config_layout.host_networking_resolv_conf]

    async def run_container(self, *args, raise_on_running=True,
                            networking=False,
                            as_pid2=True):
        async with self._operation_lock:
            if self.running:
                if raise_on_running:
                    raise RuntimeError('{} already running'.format(self))
                return self.process
            if self.cleanup_future:
                try:
                    await self.cleanup_future
                except BaseException:
                    pass
            net_args = await self.do_network_config(networking)
            if as_pid2:
                net_args.insert(0, '--as-pid2')
            if networking:
                await self.start_dependencies()
            self.cleanup_future = self.done_future()
            logger.info("Starting container {}: {}".format(
                self.name,
                " ".join(args)))
            if hasattr(self, 'model') and hasattr(self.model, 'container_args'):
                net_args = self.model.container_args + net_args
            if hasattr(self, 'container_args'):
                net_args = self.container_args + net_args
            # Move systemd options forward
            to_delete = 0
            for a in args:
                if a.startswith('--'):
                    to_delete += 1
                    net_args.insert(0, a)
                else:
                    break
            args = args[to_delete:]
            self.process = sh.systemd_nspawn("--directory=" + str(self.volume.path),
                                             '--machine=' + self.full_name,
                                             "--setenv=DEBIAN_FRONTEND=noninteractive",
                                             *net_args,
                                             *args,
                                             _bg=True,
                                             _bg_exc=False,
                                             _done=self._done_cb,
                                             _out=self._out_cb,
                                             _err_to_out=True,
                                             _tty_out=True,
                                             _in="/dev/null",
                                             _encoding='utf-8',
                                             _new_session=False,
                                             _env=self._environment(networking),
                                             )

            self.running = True
            return self.process

    async def stop_container(self):
        async with self._operation_lock:
            if not self.running:
                return
            if self.process is not None:
                self.process.terminate()
                process = self.process
                self.process = None
                try:
                    await process
                except sh.SignalException_SIGTERM:
                    pass
            else:
                await sh.machinectl("stop", self.full_name,
                                    _bg=True, _bg_exc=False)
                try:
                    await sh.ip("netns", "del", self.full_name,
                                _bg=True, _bg_exc=False)
                except BaseException:
                    pass
                self._done_cb(code=0, success=True, cmd=None)
            await super().stop_machine()

    stop_machine = stop_container

    def _done_cb(self, cmd, success, code):
        def callback():
            # Callback needed to run in IO loop thread because futures
            # do not trigger their done callbacks in a threadsafe
            # manner.
            for f in self._done_waiters:
                if not f.cancelled():
                    f.set_result(0 if success else code)
            self._done_waiters = []
            if self.network_namespace:
                self.network_namespace.close()
            self.network_namespace = None
        logger.info("Container {} exited with code {}".format(
            self.name, code))
        for k in ('shell', 'container_leader'):
            try:
                del self.__dict__[k]
            except KeyError:
                pass
        self.running = False
        self.loop.call_soon_threadsafe(callback)

    def done_future(self):
        future = self.loop.create_future()
        self._done_waiters.append(future)
        return future

    def _out_cb(self, data):
        data = data.strip()
        logger.debug("Container {}: output {}".format(self. name,
                                                      data))

        for selector in self._out_selectors:
            r, cb, once = selector
            if cb is None:
                continue
            m = r.search(data)
            if m:
                try:
                    self.loop.call_soon_threadsafe(cb, m, data)
                except Exception:
                    logger.exception("Container {}: Error calling {}".format(
                        self.name, cb))
                if once:
                    # Free the RE and callback
                    selector[0:2] = [None, None]

    def find_output(self, regexp, cb, once):
        regexp = re.compile(regexp)
        assert isinstance(once, bool)
        self._out_selectors.append([regexp, cb, once])

    async def start_container(self, *args):
        def started_callback(m, data):
            started_future.set_result(True)
        if self.running:
            return
        started_future = self.loop.create_future()
        self.find_output(r'\].*Reached target.*Basic System', started_callback, True)
        # run_container calls start_dependencies
        await super().start_machine()
        await self.run_container("--kill-signal=SIGRTMIN+3", *args, "/bin/systemd",
                                 networking=True, as_pid2=False,
                                 raise_on_running=False)
        done_future = self.done_future()
        await asyncio.wait([done_future, started_future],
                           return_when="FIRST_COMPLETED")
        if done_future.done():
            logger.error("Container {} failed to start".format(self.name))
            raise RuntimeError("Container failed to start")
        assert started_future.result() is True
        logger.info("Container {} started".format(self.name))

    start_machine = start_container

    @setup_task('ssh-keygen')
    async def generate_ssh_keys(self):
        if self.skip_ssh_keygen:
            raise SkipSetupTask
        process = await self.run_container("/usr/bin/ssh-keygen", "-A")
        await process
        self.ssh_rekeyed()

    async def container_command(self, *args, _bg=True, _bg_exc=False, **kwargs):
        '''Call :meth:`run_container` and await the resulting process.
        '''
        process = await self.run_container(*args, **kwargs)
        return await process

    def close(self, canceled_futures=None):
        if hasattr(self, 'volume'):
            if self.close_volume:
                self.volume.close()
            del self.volume
            self.injector.close(canceled_futures=canceled_futures)

    def __del__(self):
        self.close()

    async def network_online(self):
        await self.shell('/bin/systemctl', "start", "network-online.target",
                         _bg=True, _bg_exc=False
                         )

    @memoproperty
    def container_leader(self):
        return str(sh.machinectl('-pLeader', '--value', 'show', self.full_name,
                                 _in="/dev/null",
                                 _tty_out=False, _bg=False
                                 ).stdout,
                   'utf-8').strip()

    @memoproperty
    def shell(self):
        # You might think you want to use machinectl shell to create a shell.
        # That might be nice, except that you don't get exit values so you don't
        # know if your shell commands succeed or not.
        if not self.running:
            raise RuntimeError("Container not running")

        return sh.nsenter.bake("-t" + self.container_leader, "-C", "-m", "-n", "-u", "-i", "-p",
                               _env=self._environment())

    def _environment(self, networking=False):
        env = os.environ.copy()
        env['DEBIAN_FRONTEND'] = 'noninteractive'
        if networking:
            env['SYSTEMD_NSPAWN_API_VFS_WRITABLE'] = 'network'
        return env

    @contextlib.asynccontextmanager
    async def filesystem_access(self):
        yield Path(self.volume.path)

    @contextlib.asynccontextmanager
    async def ansible_not_running_context(self):
        '''
        When a :class:`Container`  that is not running is used as a host in a call to :func:`carthage.ansible.run_playbook`, then the ansible *chroot* plugin is used to connect to the container.  However, configuration is changed to use ``nsenter`` so that an actual namespace is used.
'''
        process = await self.run_container("/bin/bash")  # create a container; the bash does nothing
        try:
            yield {
                'ansible_connection': 'chroot',
                # Overriding ansible_executable is a hack; it's one of the arguments we
                # can specify and we need to pass the name to our helper
                'ansible_executable': self.full_name,
                'ansible_host': str(self.volume.path),
                'ansible_chroot_exe': str((_resources_path / "ansible-chroot-helper").absolute()),
            }
        finally:
            try:
                await self.stop_container()
            except Exception:
                pass

    def _apply_to_container_customization(self, customization):
        '''A method indicating that this object can have :class:`~carthage.machine.ContainerCustomizations` applied.  Provides a mechanism for adapting the customization if needed to a particular container-like machine.  Not needed for this class.
        '''
        return

    def _apply_to_filesystem_customization(self, customization):
        customization.path = self.volume.path
        customization.run_command = self.container_command

__all__ = ['container_volume', 'container_image', 'Container']



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import json
import logging
import os
import os.path
import shutil
import types
import mako
import mako.lookup
import mako.template
from pathlib import Path
from .dependency_injection import *
from .utils import when_needed, memoproperty
from .image import SetupTaskMixin, setup_task, ImageVolume
from .machine import Machine, SshMixin, ContainerCustomization, disk_config_from_model
from . import sh
from .config import ConfigLayout
from .ports import PortReservation
import carthage.network

logger = logging.getLogger('carthage.vm')

_resources_path = os.path.join(os.path.dirname(__file__), "resources")
_templates = mako.lookup.TemplateLookup([_resources_path + '/templates'])


vm_image = InjectionKey('vm-image')

# Our capitalization rules are kind of under-sspecified.  We're not
# upcasing all letters of acronyms in camel-case compounds, but Vm
# seems strange.  VM is canonical but Vm is an accepted alias.


@inject_autokwargs(
    config_layout=ConfigLayout,
    injector=Injector,
    image=InjectionKey(vm_image, _ready=False),
    network_config=carthage.network.NetworkConfig
)
class VM(Machine, SetupTaskMixin):

    def __init__(self, name, *, console_needed=False,
                 **kwargs):
        super().__init__(name=name, **kwargs)
        injector = self.injector
        config_layout = self.config_layout
        self.console_needed = console_needed
        if self.console_needed:
            self.console_port = injector(PortReservation)
        self.running = False
        self.closed = False
        self.volume = None
        self.vm_running = self.machine_running
        self._operation_lock = asyncio.Lock()

    def gen_volume(self):
        if self.volume is not None:
            return
        self.volume = self.image.clone_for_vm(self.name)
        self.ssh_rekeyed()
        os.makedirs(self.stamp_path, exist_ok=True)

    async def write_config(self):
        template = _templates.get_template("vm-config.mako")
        await self.resolve_networking()
        for i, link in self.network_links.items():
            await link.instantiate(carthage.network.BridgeNetwork)
            await self.image.async_become_ready()
        self.gen_volume()
        ci_data = None
        if self.model and getattr(self.model, 'cloud_init', False):
            ci_data = await self.ainjector(carthage.cloud_init.generate_cloud_init_cidata)
        disk_config = []
        async for d in await self.ainjector(qemu_disk_config, self, ci_data):
            disk_config.append(types.SimpleNamespace(**d))
        with open(self.config_path, 'wt') as f:
            f.write(template.render(
                console_needed=self.console_needed,
                console_port=self.console_port.port if self.console_needed else None,
                name=self.full_name,
                links=self.network_links,
                model_in=self.model,
                disk_config=disk_config,
                if_name=lambda n: carthage.network.base.if_name(
                    "vn", self.config_layout.container_prefix, n.name, self.name),
                volume=self.volume))
            if self.console_needed:
                with open(self.console_json_path, "wt") as f:
                    f.write(self._console_json())

    @memoproperty
    def config_path(self):
        return os.path.join(self.config_layout.vm_image_dir, self.name + '.xml')

    @memoproperty
    def console_json_path(self):
        return os.path.join(self.config_layout.vm_image_dir, self.name + '.console')

    async def start_vm(self):
        async with self._operation_lock:
            if self.running is True:
                return
            await self.start_dependencies()
            await super().start_machine()
            await self.write_config()
            await sh.virsh('create',
                           self.config_path,
                           _bg=True, _bg_exc=False)
            if self.__class__.ip_address is Machine.ip_address:
                try:
                    self.ip_address
                except NotImplementedError:
                    try:
                        await self._find_ip_address()
                    except e:
                        sh.virsh("destroy", self.full_name,
                                 _bg=True, _bg_exc=False)
                        raise e from None
            self.running = True

    start_machine = start_vm

    async def stop_vm(self):
        async with self._operation_lock:
            if not self.running:
                return

            await sh.virsh("shutdown", self.full_name,
                           _bg=True,
                           _bg_exc=False)
            for i in range(10):
                await asyncio.sleep(5)
                try:
                    sh.virsh('domid', self.full_name, _bg=False)
                except sh.ErrorReturnCode_1:
                    # it's shut down
                    self.running = False
                    break
            if self.running:
                try:
                    sh.virsh('destroy', self.full_name, _bg=False)
                except sh.ErrorReturnCode:
                    pass
                self.running = False
                await super().stop_machine()

    stop_machine = stop_vm

    def close(self, canceled_futures=None):
        if self.closed:
            return
        if (not self.config_layout.persist_local_networking) or self.config_layout.delete_volumes:
            if self.running:
                try:
                    sh.virsh("destroy", self.full_name, _bg=False)
                    self.running = False
                except Exception:
                    pass
            try:
                os.unlink(self.config_path)
            except FileNotFoundError:
                pass
        if self.config_layout.delete_volumes:
            try:
                shutil.rmtree(self.stamp_path)
            except FileNotFoundError:
                pass
        if self.volume:
            self.volume.close()
        self.injector.close(canceled_futures=canceled_futures)
        self.closed = True

    def __del__(self):
        self.close()

    async def async_ready(self):
        await self.write_config()
        await self.is_machine_running()

        await self.run_setup_tasks(context=self.machine_running(ssh_online=True))
        return await super().async_ready()

    async def is_machine_running(self):
        try:
            sh.virsh('domid', self.full_name, _bg=False)
            self.running = True
            if self.running and (self.__class__.ip_address is Machine.ip_address):
                try:
                    self.ip_address
                except NotImplementedError:
                    await self._find_ip_address()
        except sh.ErrorReturnCode_1:
            self.running = False
        return self.running

    async def _find_ip_address(self):
        for i in range(30):
            try:
                res = await sh.virsh("qemu-agent-command",
                                     self.full_name,
                                     '{"execute":"guest-network-get-interfaces"}',
                                     _bg=True, _bg_exc=False, _timeout=5)
            except sh.ErrorReturnCode_1 as e:
                # We should retry in a bit if the message contains 'not connected' and fail for other errors
                if b'connected' not in e.stderr:
                    raise
                await asyncio.sleep(3)
                continue

            js_res = json.loads(res.stdout)
            for item in js_res['return']:
                if item['name'] == 'lo':
                    continue
                if 'ip-addresses' not in item:
                    continue
                for addr in item['ip-addresses']:
                    if addr['ip-address'].startswith('fe80'):
                        continue
                    self.ip_address = addr['ip-address']
                    return
            await asyncio.sleep(3)

    @property
    def stamp_path(self):
        return Path(str(self.volume.path) + '.stamps')

    def _console_json(self):

        d = {
            "password": "aces",
            "user": "aces",
            "port": self.console_port.port,
            "host": sh.hostname('--fqdn', _encoding='utf-8', _bg=False).strip(),
            "description": self.full_name,
            "type": "spice",
            "ca": self.vm_ca(),
        }

        return json.dumps(d)

    @classmethod
    def vm_ca(cls):
        paths = ('/etc/pki/libvirt-spice', '/etc/pki/qemu')
        for p in paths:
            ca_file = os.path.join(p, 'ca-cert.pem')
            if os.path.exists(ca_file):
                with open(ca_file, 'rt') as f:
                    ca = f.read()
                ca = ca.replace("\n", "\\n")
                ca = ca.replace("\r", "")
                return ca
        raise FileNotFoundError


Vm = VM


class InstallQemuAgent(ContainerCustomization):

    description = "Install qemu guest agent"

    @setup_task("Install qemu guest agent")
    async def install_guest_agent(self):
        await self.container_command("/usr/bin/apt", "-y", "install", "qemu-guest-agent")


@inject(ainjector=AsyncInjector)
async def qemu_disk_config(vm, ci_data, *, ainjector):
    # Handle qemu specific disk_config
    disk_config = disk_config_from_model(getattr(vm, 'model', {}),
                                         default_disk_config=[
                                             dict(),
                                             dict(target_type='cdrom',
                                                  source_type='file',
                                                  driver='raw',
                                                  qemu_source='file',
                                                  readonly=True)],
                                         )

    for i, entry in enumerate(disk_config):
        if i == 0:  # primary disk
            if 'volume' not in entry:
                entry['volume'] = vm.volume
        if 'cache' not in entry:
            try:
                entry['cache'] = vm.model.disk_cache
            except AttributeError:
                entry['cache'] = 'writethrough'
        if 'volume' not in entry and 'size' in entry:
            entry['volume'] = await ainjector(
                ImageVolume, name=vm.name + f'_disk_{i}',
                create_size=entry.size,
                unpack=False)
        try:
            await entry['volume'].async_become_ready()
        except AttributeError:
            pass  # QemuVolume is not AsyncInjectable
        except KeyError:
            pass  # volume not set
        entry.setdefault('target_type', 'disk')
        if 'volume' in entry:
            entry.update(entry['volume'].qemu_config(entry))
        entry.setdefault('source_type', 'file')
        entry.setdefault('driver', 'raw')
        entry.setdefault('qemu_source', 'dev' if entry['source_type'] == 'block' else 'file')
        yield entry
    if ci_data:
        yield dict(
            target_type='cdrom',
            source_type='file',
            qemu_source='file',
            driver='raw',
            readonly=True,
            path=ci_data,
            cache='writeback')

        __all__ = ('VM', 'Vm', 'InstallQemuAgent')



# Copyright (C) 2018, 2019, 2020, 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import dependency_injection
from carthage.dependency_injection import *
from carthage.dependency_injection.introspection import *
from carthage.utils import when_needed
from carthage.pytest import async_test
from carthage_test_utils import Trigger

import asyncio
import pytest


@pytest.fixture()
def injector():
    injector = dependency_injection.Injector()
    injector.add_provider(asyncio.get_event_loop(), close=False)
    return injector


@pytest.fixture()
def a_injector(injector, loop):
    a_injector = injector(dependency_injection.AsyncInjector, loop=loop)
    yield a_injector
    a_injector.close()


def test_injector_provides_self(injector):
    @inject(i=dependency_injection.Injector)
    def func(i):
        return i
    assert isinstance(injector(func), dependency_injection.Injector)


def test_injector_available(injector):
    assert isinstance(injector, dependency_injection.Injector)


def test_override_dependency(injector):
    k = dependency_injection.InjectionKey('some key')
    injector.add_provider(k, 30)

    @inject(arg=k)
    def func(arg):
        assert arg == 20
    injector(func, arg=20)
    # And make sure without the override the injector still provides the right thing

    @inject(i=k)
    def func2(i):
        assert i == 30
    injector(func2)


def test_override_replaces_subinjector(injector):
    class OverrideType:
        pass
    o1 = OverrideType()
    o2 = OverrideType()
    assert o1 is not o2

    @inject(o=OverrideType,
            i=dependency_injection.Injector)
    def func(i, o):
        assert o is o2
        assert injector is not i
        assert i.parent_injector is injector

    @inject(o=OverrideType)
    def func2(o):
        assert o is o1
    injector.add_provider(o1)
    injector(func, o=o2)
    injector(func2)


def test_injector_instantiates(injector):
    class SomeClass(dependency_injection.Injectable):
        pass

    @inject(s=SomeClass)
    def func(s):
        assert isinstance(s, SomeClass)
    injector.add_provider(SomeClass)
    injector(func)


def test_async_injector_construction(loop, injector):
    @inject(a=dependency_injection.AsyncInjector)
    def f(a):
        assert isinstance(a, dependency_injection.AsyncInjector)

    injector(f)


@async_test
async def test_construct_using_coro(a_injector, loop):
    async def coro():
        return 42
    k = dependency_injection.InjectionKey('run_coro')

    @inject(v=k)
    def f(v):
        assert v == 42
    a_injector.add_provider(k, coro)
    await a_injector(f)


@async_test
async def test_async_function(a_injector, loop):
    class Dependency(dependency_injection.Injectable):
        pass

    async def setup_dependency(): return Dependency()
    called = False

    @inject(d=Dependency)
    async def coro(d):
        assert isinstance(d, Dependency)
        nonlocal called
        called = True
    a_injector.add_provider(InjectionKey(Dependency), setup_dependency)
    await a_injector(coro)
    assert called is True


@async_test
async def test_async_ready(a_injector, loop):
    class AsyncDependency(dependency_injection.AsyncInjectable):
        async def async_ready(self):
            self.ready = True
            await super().async_ready()
            return self

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.ready = False

    @inject(r=AsyncDependency)
    def is_ready(r):
        assert r.ready
    await a_injector(is_ready, r=AsyncDependency)


def test_allow_multiple(injector):
    from carthage.config import ConfigLayout
    injector.add_provider(ConfigLayout, allow_multiple=True)
    s1 = dependency_injection.Injector(injector)
    s2 = dependency_injection.Injector(injector)
    assert s1 is not s2
    assert s1.parent_injector is injector
    assert s2.parent_injector is injector
    c1 = s1.get_instance(ConfigLayout)
    c2 = s2.get_instance(ConfigLayout)
    assert isinstance(c1, ConfigLayout)
    assert isinstance(c2, ConfigLayout)
    assert c1 is not c2
    c3 = injector.get_instance(ConfigLayout)
    assert isinstance(c3, ConfigLayout)
    assert c3 is not c1
    assert c3 is not c2


def test_allow_multiple_provider_at_root(injector):
    from carthage.config import ConfigLayout
    injector.add_provider(ConfigLayout, allow_multiple=True)
    s1 = dependency_injection.Injector(injector)
    s2 = dependency_injection.Injector(injector)
    assert s1 is not s2
    c3 = injector.get_instance(ConfigLayout)
    c1 = s1.get_instance(ConfigLayout)
    c2 = s2.get_instance(ConfigLayout)
    assert c3 is c1
    assert c2 is c3


def test_allow_multiple_false(injector):
    from carthage.config import ConfigLayout
    injector.add_provider(ConfigLayout, allow_multiple=False)
    s1 = dependency_injection.Injector(injector)
    s2 = dependency_injection.Injector(injector)
    assert s1 is not s2
    c1 = s1.get_instance(ConfigLayout)
    c2 = s2.get_instance(ConfigLayout)
    assert c1 is c2


@async_test
async def test_when_needed(a_injector, loop):
    class foo(dependency_injection.Injectable):

        def __init__(self):
            nonlocal called
            assert called is False
            called = True

    wn = when_needed(foo)
    i1 = InjectionKey('i1')
    i2 = InjectionKey('i2')
    called = False
    a_injector.add_provider(i1, wn)
    a_injector.add_provider(i2, wn)
    i1r = await a_injector.get_instance_async(i1)
    assert isinstance(i1r, foo)
    i2r = await a_injector.get_instance_async(i2)
    assert i2r is i1r
    assert await a_injector(wn) is i1r
    assert called


@async_test
async def test_when_needed_override(a_injector, loop):
    k = dependency_injection.InjectionKey('foo')
    a_injector.add_provider(k, 20)

    @dependency_injection.inject(n=k)
    def func(n):
        assert n == 29
        return "foo"
    wn = when_needed(func, n=29)
    assert await a_injector(wn) == "foo"


@async_test
async def test_when_needed_cancels(loop, a_injector):
    injector = await a_injector(dependency_injection.Injector)
    ainjector = injector(dependency_injection.AsyncInjector)
    cancelled = False
    k = dependency_injection.InjectionKey("bar")

    async def func():
        nonlocal cancelled
        try:
            await asyncio.sleep(0.5)
        except asyncio.CancelledError:
            cancelled = True
        return 39
    ainjector.add_provider(k, when_needed(func))
    loop.create_task(ainjector.get_instance_async(k))
    await asyncio.sleep(0.1)
    await dependency_injection.shutdown_injector(ainjector)
    assert cancelled is True


def test_injectable_sets_dependencies(injector):
    "Test that the constructor for Injectable tries to store dependencies as instance variables"
    k = InjectionKey("test_key")

    @inject(foo=k)
    class i1(dependency_injection.Injectable):
        pass
    injector.add_provider(k, 33)
    i1_obj = injector(i1)
    assert i1_obj.foo == 33


def test_injectable_fails_on_unknown_args(injector):
    k = InjectionKey("test_key")

    @inject(foo=k)
    class i1(dependency_injection.Injectable):
        pass
    injector.add_provider(k, 33)
    with pytest.raises(TypeError):
        i1_obj = injector(i1, bar=40)


def test_injectable_autokwargs(injector):
    k = InjectionKey("test_key")

    @inject_autokwargs(foo=k)
    class i1(dependency_injection.Injectable):
        pass
    injector.add_provider(k, 40)
    i1_obj = injector(i1)
    assert i1_obj.foo == 40
    with pytest.raises(TypeError):
        i1()


@async_test
async def test_injectable_inheritance(injector, a_injector):
    from carthage.dependency_injection import Injectable
    k1 = InjectionKey("k1")
    k2 = InjectionKey("k2")

    @inject_autokwargs(k1=k1)
    class a(Injectable):
        pass

    @inject(k2=k2)
    class b(a):
        pass

    @inject(k1=None)
    class c(b):

        def __init__(self, **kwargs):
            super().__init__(k1=20, **kwargs)

    injector.add_provider(k1, 10)
    injector.add_provider(k2, 30)
    b_obj = injector(b)
    assert b_obj.k1 == 10
    assert b_obj.k2 == 30
    c_obj = injector(c)
    assert c_obj.k2 == 30
    assert c_obj.k1 == 20

    a_injector.add_provider(k1, 30)
    a_injector.add_provider(k2, 55)
    c_obj2 = await a_injector(c)
    assert c_obj2.k1 == 20


@async_test
async def test_injector_claiming(injector, a_injector):
    ainjector = a_injector
    i2 = injector(dependency_injection.Injector)
    assert i2.claimed_by is None
    assert i2.claim() is i2
    i3 = i2.claim()
    assert i2 is not i3
    assert i3.claimed_by

    class c(AsyncInjectable):
        pass
    c_obj = await ainjector(c)
    assert c_obj.injector.claimed_by() is c_obj
    ai2 = c_obj.ainjector.claim()
    assert ai2 is not c_obj.ainjector
    assert ai2.injector is not c_obj.ainjector.injector
    assert c_obj.ainjector.injector is c_obj.injector
    # If you are permitted to override injectors on a call to an
    # injector, interesting semantic questions come up; are you just
    # overriding the kwarg, or are you also overriding what the
    # subinjector will provide when asked to provide an injector.  If
    # so, you are violating the invarient that injectors always inject
    # themselves when asked for an injector.  In any case, if it ever
    # becomes possible to override the injector keyword, the following
    # test probably should pass for any reasonable semantics.
    with pytest.raises(dependency_injection.ExistingProvider):
        c2_obj = await ainjector(c, injector=i3)
        assert c2_obj.injector.claimed_by() is c2_obj
        assert c2_obj.injector.parent_injector is i3


def test_injection_key_copy():
    i1 = InjectionKey(int)
    assert i1 is InjectionKey(i1)
    i2 = InjectionKey(i1, optional=True)
    assert i2.optional
    assert i2 == i1
    assert i2 is not i1


def test_none_kwargs():
    class foo(dependency_injection.Injectable):
        pass
    injector = dependency_injection.Injector()
    injector.add_provider(foo)
    @inject(i=foo)
    def isinst(i): return isinstance(i, foo)
    assert injector(isinst) is True
    assert injector(isinst, i=None) is False


@async_test
async def test_async_not_ready(a_injector):
    class AsyncDependency(AsyncInjectable):
        ready = False

        async def async_ready(self):
            await super().async_ready()
            self.ready = True

    @inject(baz=InjectionKey(AsyncDependency, _ready=True))
    class AsyncDependency2(AsyncDependency):
        pass
    k = InjectionKey("baz")
    k2 = InjectionKey("bazquux")
    a_injector.add_provider(k, AsyncDependency)
    a_injector.add_provider(k2, AsyncDependency2)
    nr = await a_injector.get_instance_async(InjectionKey(k, _ready=False))
    assert isinstance(nr, AsyncDependency)
    assert nr.ready is False
    nr2 = await a_injector.get_instance_async(k)
    assert nr2 is nr
    assert nr.ready is True
    nr3 = await a_injector.get_instance_async(InjectionKey(k2, _ready=False))
    assert nr3.baz.ready is True
    assert nr3.ready is False


@async_test
async def test_dependency_quote(a_injector):
    class AsyncDependency(AsyncInjectable):

        async def async_ready(self):
            raise AssertionError

    k = InjectionKey("foo")

    @inject(foo=k)
    def func(foo):
        assert foo is AsyncDependency
    await a_injector(func, foo=dependency_injection.dependency_quote(AsyncDependency))


def test_injector_xref(injector):
    injector_xref = dependency_injection.injector_xref

    class Target(Injectable):
        pass

    @inject(injector=dependency_injection.Injector)
    class Sub(Injectable):

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.injector.add_provider(InjectionKey(42), Target)
    injector.add_provider(Sub)
    injector.add_provider(InjectionKey("target"),
                          injector_xref(
                              InjectionKey(Sub),
                              InjectionKey(Target)))
    res = injector.get_instance(InjectionKey("target"))
    assert isinstance(res, Target)


def test_system_dependency_injection_keys(injector):
    "Test that SystemDependencies inject with a name"
    from carthage.system_dependency import MachineDependency, SystemDependency
    md = MachineDependency("foo.com")
    injector.add_provider(md)
    md2 = injector.get_instance(InjectionKey(SystemDependency, name="foo.com"))
    assert md is md2


@async_test
async def test_concurrent_resolution(a_injector, loop):
    "Test that if a  get_instance has already recorded a partial future, kwarg placement ends up working"
    ainjector = a_injector
    fut_inprogress = loop.create_future()
    fut_unblock = loop.create_future()

    async def block_for_a_while():
        fut_inprogress.set_result(True)
        await fut_unblock
        return 333

    key = InjectionKey("async_function")
    ainjector.add_provider(key, block_for_a_while)

    @inject(foo=key)
    async def wait_for_foo(foo):
        assert foo == 333
        return True

    fut_key = loop.create_task(ainjector.get_instance_async(key))
    # at this point we've requested key to be resolved, but probably block_for_a_while has not even started
    await fut_inprogress
    # At this point block_for_a_while is running, and that future has been recorded as the provider for key
    fut_wait_for_foo = loop.create_task(ainjector(wait_for_foo))
    # Now wait a bit so that wait_for_foo is blocking on the result of key
    # If the time becomes a problem we could probably have a second dependency
    # for wait_for_foo that sets its own future and wait for that here
    await asyncio.sleep(0.2)
    fut_unblock.set_result(True)
    assert await fut_wait_for_foo is True
    await fut_key


@async_test
async def test_async_become_ready_handles_dependencies(a_injector):
    ainjector = a_injector

    @inject(bar=InjectionKey("bar"))
    class Foo(AsyncInjectable):

        async def async_ready(self):
            assert self.bar.ready is True
            return await super().async_ready()

    class Bar(AsyncInjectable):

        ready = False

        async def async_ready(self):
            self.ready = True
            return await super().async_ready()

    ainjector.add_provider(InjectionKey("bar"), Bar)
    ainjector.add_provider(Foo)
    foo = await ainjector.get_instance_async(InjectionKey(Foo, _ready=False))
    assert foo.bar.ready is False
    await foo.async_become_ready()
    assert foo.bar.ready is True


@async_test
async def test_notready_cycles_okay(a_injector):
    ainjector = a_injector

    class Cycle(AsyncInjectable):

        async def async_ready(self):
            # If things fail this will hang
            res = await self.ainjector.filter_instantiate_async(Cycle, ['name'], ready=False)
            assert res[0][1] is self
            return await super().async_ready()

    foo_key = InjectionKey(Cycle, name="foo")
    ainjector.add_provider(foo_key, Cycle)
    await ainjector.get_instance_async(foo_key)


@async_test
async def test_introspection_registers_root(a_injector, loop):
    ainjector = a_injector

    class c(AsyncInjectable):
        async def async_ready(self):
            trigger_1.trigger()
            await trigger_2
            await super().async_ready()

    with Trigger() as trigger_1, Trigger() as trigger_2:
        assert len(instantiation_roots) == 0
        ainjector.add_provider(c)
        c_future = loop.create_task(ainjector.get_instance_async(c))
        await trigger_1
        assert len(instantiation_roots) == 1
        for ic in instantiation_roots:
            assert hasattr(ic, 'key')
            assert ic.key is InjectionKey(c)
            assert len(ic.dependencies_waiting) == 0
        trigger_2.trigger()
        await c_future
    assert len(instantiation_roots) == 0


@async_test
async def test_ready_event_propagation(a_injector, loop):
    "Make sure that when multiple instantiations of an object, one to ready and one not to ready are in progress, the right context is recorded in dependency_waiting so that we can subscribe to the right events."
    ainjector = a_injector

    class ToReady(AsyncInjectable):

        async def async_resolve(self):
            trigger_async_resolve_started.trigger()
            await trigger_async_resolve
            return self

        async def async_ready(self):
            trigger_async_ready_started.trigger()
            return await super().async_ready()

    @inject(to_ready=InjectionKey(ToReady, _ready=False))
    class Dependent1(Injectable):
        pass

    @inject(to_ready=ToReady)
    class Dependent2(Injectable):
        pass

    def dp_callback(*args, target, **kwargs):
        if target.ready:
            trigger_event_received.trigger()
    sub_injector = await ainjector(Injector)
    sub_ainjector = sub_injector(AsyncInjector)
    # we start in the super injector with something that will cause
    # ToReady to get started being instantiated.  In the sub injector
    # we pick that up as an in-progress dependency and make sure that
    # if we look up the dependency in the introspection interface and
    # subscribe to that, we get the right callback.  I.E. we make sure
    # we get the InstantiationContext waiting in the higher level
    # injector.
    with Trigger() as trigger_async_resolve_started, \
            Trigger() as trigger_async_resolve, \
            Trigger() as trigger_async_ready_started, \
            Trigger() as trigger_event_received:
        ainjector.add_provider(Dependent1)
        ainjector.add_provider(ToReady)
        assert len(instantiation_roots) == 0
        dependent1_future = loop.create_task(
            ainjector.get_instance_async(Dependent1))
        await trigger_async_resolve_started
        sub_ainjector.add_provider(Dependent2)
        dependent2_future = loop.create_task(
            sub_ainjector.get_instance_async(Dependent2))
        trigger_async_resolve.trigger()
        with ainjector.injector.event_listener_context(
                InjectionKey(ToReady), "dependency_final", dp_callback) as futures:
            await dependent2_future
        trigger_async_ready_started.assert_triggered()
        if futures:
            await asyncio.gather(futures)
        trigger_event_received.assert_triggered()
    assert len(instantiation_roots) == 0


def test_optional_not_present(injector):
    class SomeDependency(Injectable):
        pass

    @inject(dep=InjectionKey(SomeDependency, _optional=NotPresent))
    def func(**kwargs):
        assert len(kwargs) == 0
    injector(func)

    @inject_autokwargs(dep=InjectionKey("foo", _optional=NotPresent))
    class AnotherDependency(Injectable):
        pass
    instance = injector(AnotherDependency)
    assert not hasattr(instance, 'dep')
    injector.add_provider(InjectionKey('foo'), 42)
    instance2 = injector(AnotherDependency)
    assert instance2.dep == 42
@async_test
async def test_resolve_deferred(ainjector):
    foo = InjectionKey('foo')
    bar = InjectionKey('bar')
    baz = InjectionKey('baz')
    @inject(baz=baz)
    def func(baz, quux):
        return baz*quux
    args = dict(quux=10)
    ainjector.add_provider(foo, 'foo')
    ainjector.add_provider(bar, 'bar')
    ainjector.add_provider(baz, 4)
    result = await ainjector(
        resolve_deferred, ainjector, [
        90,
        dict(a=10,b=20),
        dict(c=foo,d=func),
        bar], args=args)
    assert result == [
        90,
        dict(a=10, b=20),
        dict(c='foo', d=40),
        'bar']



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from . import base
from .base import *
from .base import external_network  # Not really sure it should be in all
__all__ = base.__all__

from .mac import random_mac_addr, MacStore, persistent_random_mac, persistent_random_mac_always

__all__ += ['random_mac_addr', 'MacStore', 'persistent_random_mac', 'persistent_random_mac_always']

from .namespace import NetworkNamespace
__all__ += ['NetworkNamespace']

from .config import V4Config
__all__ += ['V4Config']
